import os
import copy
import time
from functools import wraps
import warnings
from typing import Tuple, Optional, Callable, Any, Dict, List, Type, Union

import torch.nn.functional as f
from torch.optim import Optimizer
import torch
from optuna.trial._base import BaseTrial
import numpy as np
from tqdm import tqdm
from sklearn.model_selection._split import BaseCrossValidator
from sklearn.model_selection import KFold, train_test_split
from torch.utils.data.sampler import SubsetRandomSampler
from torch.utils.data import DataLoader
import optuna
from datetime import datetime
from torch.utils.tensorboard.writer import SummaryWriter
from torch.optim.lr_scheduler import _LRScheduler  # noqa

from ..utility.optimization import MissingClosureError
from gdeep.models import ModelExtractor
from gdeep.utility import _inner_refactor_scalars
from gdeep.utility import DEVICE
from .metrics import accuracy

from gdeep.utility.custom_types import Tensor

try:
    import torch_xla.core.xla_model as xm  # type: ignore
    import torch_xla.distributed.xla_multiprocessing as xmp  # type: ignore
    import torch_xla.distributed.parallel_loader as pl  # type: ignore

    try:
        DEVICE = xm.xla_device()
    except NameError:
        pass
    print("Using TPU!")
except ModuleNotFoundError:
    print("No TPUs...")


def _add_data_to_tb(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """decorator to store PR data to tensorboard"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        pred, val_loss, correct = func(*args, **kwargs)
        try:
            # add data to tensorboard
            Trainer._add_pr_curve_tb(
                torch.vstack(pred),
                kwargs["class_label"],
                kwargs["class_probs"],
                kwargs["writer_tag"] + "/validation",
            )
        except NotImplementedError:
            warnings.warn("The PR curve is not being filled because too few data exist")
        return pred, val_loss, correct

    return wrapper


class Trainer:
    """This is the generic class that allows
    the user to benchmark models over architectures
    datasets, regularisation, metrics... in one line
    of code.

    Args:
        model :
            standard torch model
        dataloaders (list of utils.DataLoader):
            list of standard torch DataLoaders, e.g.
            `[dl_tr, dl_val, dl_ts]`
        loss_fn :
            loss function to average over batches
        writer :
            tensorboard writer
        training_metric:
            the function that computes the metric: it shall
            have two arguments, one for the prediction
            and the other for the ground truth
        k_fold_class (sklearn.model_selection, default ``KFold(5, shuffle=True)``):
            the class instance to implement the KFold, can be
            any of the Splitter classes of sklearn. More
            info at https://scikit-learn.org/stable/modules/classes.html#module-sklearn.model_selection

    Examples::

        from torch import nn
        from torch.optim import SGD
        from sklearn.model_selection import StratifiedKFold
        from gdeep.models import FFNet
        from gdeep.trainer import Trainer
        from gdeep.data import BuildDatasets, BuildDataLoaders
        from gdeep.search import GiottoSummaryWriter
        # model
        class model1(nn.Module):
            def __init__(self):
                super(model1, self).__init__()
                self.seqmodel = nn.Sequential(nn.Flatten(), FFNet(arch=[3, 5, 10, 5, 2]))
            def forward(self, x):
                return self.seqmodel(x)

        model = model1()
        # dataloaders
        bd = BuildDatasets(name="DoubleTori")
        ds_tr, ds_val, _ = bd.build_datasets()
        dl = BuildDataLoaders((ds_tr, ds_val))
        dl_tr, dl_val, dl_ts = dl.build_dataloaders(batch_size=23)

        # loss function
        loss_fn = nn.CrossEntropyLoss()
        # tb writer
        writer = GiottoSummaryWriter()
        # pipeline
        pipe = Trainer(model, [dl_tr, dl_val, dl_ts],
                        loss_fn, writer, None,
                        StratifiedKFold(5, shuffle=True))
        # then one needs to train the model using the pipeline!
        pipe.train(SGD, 2, True, {"lr": 0.001}, n_accumulated_grads=5)

    """

    scheduler: Optional[_LRScheduler]
    writer: Optional[SummaryWriter]
    registered_hook: Optional[
        Callable[[int, Optimizer, ModelExtractor, Optional[SummaryWriter]], Any]
    ] = None

    def __init__(
        self,
        model: torch.nn.Module,
        dataloaders: List[DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]]],
        loss_fn: Callable[[Tensor, Tensor], Tensor],
        writer: Optional[SummaryWriter] = None,
        training_metric: Optional[Callable[[Tensor, Tensor], float]] = None,
        k_fold_class: Optional[BaseCrossValidator] = None,
    ) -> None:
        self.model = model
        self.initial_model = copy.deepcopy(self.model)
        assert 0 < len(dataloaders) < 4, "Length of dataloaders must be 1, 2, or 3"
        self.dataloaders = dataloaders  # train and test
        self.train_epoch = 0
        self.val_epoch = 0
        self.best_val_loss = np.inf
        self.best_val_acc = 0.0
        self.loss_fn = loss_fn
        if training_metric:
            self.training_metric = training_metric
        else:
            self.training_metric = accuracy
        # integrate tensorboard
        self.writer = writer
        if self.writer is None:
            warnings.warn("No writer detected")
        self.check_has_trained: bool = False
        # used in gradient clipping
        self.clip: int = 5
        # used by hpo:
        self.run_name: Optional[str] = None
        self.val_loss_list_hparam: List[List[float]] = []
        self.val_acc_list_hparam: List[List[float]] = []
        self.best_not_last: bool = False
        # profiler
        self.prof: Any = None

        if not k_fold_class:
            self.k_fold_class = KFold(5, shuffle=True)
        else:
            self.k_fold_class = k_fold_class

    def _set_initial_model(self) -> None:
        """This private method is used to set
        the initial_model"""
        self.initial_model = copy.deepcopy(self.model)

    def _reset_model(self) -> None:
        """Private method to reset the initial model weights.
        This function is essential for the cross-validation
        procedure.
        """
        self.model = copy.deepcopy(self.initial_model)

    def _optimisation_step(
        self,
        steps: int,
        loss: Tensor,
        epoch_loss: float,
        batch_metric: float,
        batch: int,
        closure: Callable[[], Tensor],
    ) -> float:
        """Backpropagation"""
        if self.n_accumulated_grads < 2:  # usual case for stochastic gradient descent
            self.optimizer.zero_grad()
            loss.backward()
            if DEVICE.type == "xla":
                xm.optimizer_step(
                    self.optimizer, barrier=True
                )  # Note: Cloud TPU-specific code!
            else:
                try:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.clip)  # type: ignore
                    self.optimizer.step()
                except (MissingClosureError,):
                    self.optimizer.step(closure)  # type: ignore

        else:  # if we use gradient accumulation techniques
            (loss / self.n_accumulated_grads).backward()
            if (
                batch + 1
            ) % self.n_accumulated_grads == 0:  # do the optimization step only after the accumulations
                if DEVICE.type == "xla":
                    xm.optimizer_step(
                        self.optimizer, barrier=True
                    )  # Note: Cloud TPU-specific code!
                else:
                    try:
                        torch.nn.utils.clip_grad_norm_(  # type: ignore
                            self.model.parameters(), self.clip
                        )
                        self.optimizer.step()
                    except (MissingClosureError,):
                        self.optimizer.step(closure)  # type: ignore
                self.optimizer.zero_grad()
        if batch % 1 == 0:
            epoch_loss += loss.item()
            print(
                f"Batch training loss:  {epoch_loss / (batch + 1)}",
                f" \tBatch training {self.training_metric.__name__}: ",
                batch_metric,
                " \t[",
                batch + 1,
                "/",
                steps,
                "]                     ",
                end="\r",
            )
        return epoch_loss

    def _send_to_device(
        self, x: Union[Tensor, List[Tensor]], y: Tensor
    ) -> Tuple[Tensor, Union[Tensor, List[Tensor]], Tensor]:
        """use this private method to send the
        ``x`` and ``y`` to the ``DEVICE``.

        Args:
            x:
                the input of the model, either a List[Tensor] or a Tensor
            y:
                the label

        Returns:
            (Tensor, Union[Tensor, List[Tensor]], Tensor)
                the prediction for x, x and the label

        """
        new_x: List[Tensor] = []
        if isinstance(x, tuple) or isinstance(x, list):
            for xi in x:
                new_x.append(xi.to(DEVICE))
            x = new_x
            prediction = self.model(*x)
            if hasattr(prediction, "logits"):  # unwrapper for HuggingFace BERT model
                prediction = prediction.logits  # unwrapper for HuggingFace BERT model
        else:
            x = x.to(DEVICE)
            prediction = self.model(x)
            if hasattr(prediction, "logits"):  # unwrapper for HuggingFace BERT model
                prediction = prediction.logits  # unwrapper for HuggingFace BERT model
        y = y.to(DEVICE)

        return prediction, x, y

    def _inner_train_loop(
        self,
        dl_tr: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        writer_tag: str,
        size: int,
        steps: int,
    ) -> Tuple[float, float]:
        """Private method to run the loop
        over the batches for the optimisation"""

        if self.prof is not None:
            self.prof.start()
        metric_list = []
        epoch_loss = 0.0
        for batch, (X, y) in enumerate(dl_tr):

            def closure() -> Tensor:
                loss2 = self.loss_fn(self.model(X), y)
                loss2.backward()
                return loss2

            pred, X, y = self._send_to_device(X, y)
            batch_metric = self.training_metric(pred, y)
            metric_list.append(batch_metric)
            loss = self.loss_fn(pred, y)
            # Save to tensorboard
            try:
                self.writer.add_scalar(  # type: ignore
                    writer_tag + "/loss/train",
                    loss.item(),
                    self.train_epoch * len(dl_tr) + batch,
                )

                try:
                    top2_pred = torch.topk(pred, 2, -1).values
                    try:
                        self.writer.add_histogram(  # type: ignore
                            writer_tag + "/predictions/train",
                            torch.abs(torch.diff(top2_pred, dim=-1)),
                            self.train_epoch * steps + batch,
                        )
                    except ValueError:
                        warnings.warn(
                            f"The histogram is empty, most likely because your loss"
                            f" is exploding. Try use gradient clipping."
                        )
                except RuntimeError:
                    pass
            except AttributeError:
                pass
            epoch_loss = self._optimisation_step(
                steps, loss, epoch_loss, batch_metric, batch, closure
            )

            if self.prof is not None:
                self.prof.step()

        if self.prof is not None:
            self.prof.stop()

        # epoch metric and loss:
        epoch_metric = sum(metric_list) / len(metric_list)
        epoch_loss /= steps
        return epoch_metric, epoch_loss

    def _train_loop(
        self,
        dl_tr: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        writer_tag: str = "",
    ) -> Tuple[float, float]:
        """private method to run a single training
        loop
        """
        self.model = self.model.to(DEVICE)
        self.model.train()
        try:
            length: int = len(dl_tr.sampler.indices)  # type: ignore
        except AttributeError:
            length: int = len(dl_tr.dataset)  # type: ignore
        steps = len(dl_tr)
        tik = time.time()
        assert self.n_accumulated_grads <= steps, (
            "The number of" + " accumulated gradients shall be diminished!"
        )
        epoch_metric, epoch_loss = self._inner_train_loop(
            dl_tr, writer_tag, length, steps
        )
        print(
            f"Epoch training loss: {epoch_loss:>8f} \tEpoch training "
            f"{self.training_metric.__name__}: {epoch_metric:.2f}% ".ljust(100)
        )
        try:
            self.writer.flush()  # type: ignore
        except AttributeError:
            pass
        print(f"Time taken for this epoch: {round(time.time() - tik):.2f}s")
        try:
            print(f"Learning rate value: {self.optimizer.param_groups[0]['lr']:0.8f}")
        except KeyError:
            pass
        return epoch_metric, epoch_loss

    def _val_loop(
        self,
        dl_val: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        writer_tag: str = "",
    ) -> Tuple[float, float]:
        """private method to run a single validation
        loop
        """
        self.model = self.model.to(DEVICE)
        try:
            size = len(dl_val.sampler.indices)  # type: ignore
        except AttributeError:
            size = len(dl_val.dataset)  # type: ignore
        class_label: List[Tensor] = []
        class_probs: List[List[Tensor]] = []
        self.model.eval()

        pred_list, epoch_loss, epoch_metric = self._inner_loop(  # type: ignore
            dl=dl_val,  # type: ignore
            class_probs=class_probs,  # type: ignore
            class_label=class_label,  # type: ignore
            writer_tag=writer_tag,  # type: ignore
        )
        # accuracy
        try:
            if not self.run_name:
                self.writer.add_scalar(  # type: ignore
                    writer_tag + "/metric/validation", epoch_metric, self.val_epoch
                )
            else:
                self.val_acc_list_hparam.append([epoch_metric, self.val_epoch])
                self.val_loss_list_hparam.append([epoch_loss, self.val_epoch])

            try:
                top2_pred = torch.topk(torch.vstack(pred_list), 2, -1).values
                try:
                    self.writer.add_histogram(  # type: ignore
                        writer_tag + "/predictions/validation",
                        torch.abs(torch.diff(top2_pred, dim=-1)),
                        self.val_epoch,
                    )
                except ValueError:
                    warnings.warn(
                        f"The histogram is empty, most likely because your loss"
                        f" is exploding. Try use gradient clipping."
                    )
            except RuntimeError:
                pass
        except AttributeError:
            pass
        print(
            f"Validation results: \n {self.training_metric.__name__}: {epoch_metric:.2f}%, \
                Avg loss: {epoch_loss:>8f} \n"
        )
        try:
            self.writer.flush()  # type: ignore
        except AttributeError:
            pass

        return epoch_loss, epoch_metric

    @staticmethod
    def _add_pr_curve_tb(
        pred: Tensor,
        class_label: List[Tensor],
        class_probs: List[List[Tensor]],
        writer: SummaryWriter,
        writer_tag: str = "",
    ) -> None:
        """private function to add the PR curve
        to tensorboard"""
        probs = torch.cat([torch.stack(batch) for batch in class_probs]).cpu()
        labels = torch.cat(class_label).cpu()
        for class_index in range(len(pred[0])):
            tensorboard_truth = 1 * (labels == class_index).flatten()
            tensorboard_probs = probs[:, class_index]
            # print(tensorboard_truth)
            # print(tensorboard_probs)
            try:
                writer.add_pr_curve(
                    writer_tag + "/class = " + str(class_index),
                    tensorboard_truth,
                    tensorboard_probs,
                    global_step=0,
                )
            except AttributeError:
                warnings.warn("Cannot store data in the PR curve")
            except ValueError:
                warnings.warn("Cannot store data in the PR curve")

    @_add_data_to_tb  # type: ignore
    def _inner_loop(
        self,
        *,
        dl: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        class_probs: List[List[Tensor]],
        class_label: List[Tensor],
        writer_tag: str,  # noqa
    ) -> Tuple[List[Tensor], float, float]:
        """private function used inside the test
        and validation loops"""
        pred_list = []
        batch_metric_list = []
        loss = 0.0
        with torch.no_grad():
            for X, y in dl:
                pred, X, y = self._send_to_device(X, y)
                pred_list.append(pred)
                class_probs_batch = [f.softmax(el, dim=0) for el in pred]
                class_probs.append(class_probs_batch)
                loss += self.loss_fn(pred, y).item()
                batch_metric = self.training_metric(pred, y)
                batch_metric_list.append(batch_metric)
                class_label.append(y)
        epoch_metric = sum(batch_metric_list) / len(batch_metric_list)
        epoch_loss = loss / len(batch_metric_list)
        return pred_list, epoch_loss, epoch_metric

    def _init_profiler(
        self, profiling: bool, cross_validation: bool, n_epochs: int, k_folds: int
    ) -> None:
        """initialise the profler for profiling"""
        # profiling
        active: int = 10
        if not cross_validation:
            active = n_epochs - 2
        else:
            active = k_folds * (n_epochs - 2)

        if profiling:
            try:
                self.prof = torch.profiler.profile(  # type: ignore
                    activities=[
                        torch.profiler.ProfilerActivity.CPU,  # type: ignore
                        torch.profiler.ProfilerActivity.CUDA,  # type: ignore
                    ],
                    schedule=torch.profiler.schedule(  # type: ignore
                        wait=1, warmup=1, active=active, repeat=1
                    ),
                    on_trace_ready=torch.profiler.tensorboard_trace_handler(  # type: ignore
                        os.path.join(
                            ".",
                            "runs",
                            (
                                self.model.__class__.__name__ + str(datetime.today())
                            ).replace(":", "-"),
                        ),
                        worker_name="worker",
                    ),
                    record_shapes=True,
                    profile_memory=True
                    # with_stack=True
                )
            except AssertionError:
                pass

    def _init_optimizer_and_scheduler(
        self,
        keep_training: bool,
        cross_validation: bool,
        optimizer: Type[Optimizer],
        optimizers_param: Dict[str, Any],
        lr_scheduler: Optional[Type[_LRScheduler]] = None,
        scheduler_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Reset or maintain the LR scheduler and the
        optimizer depending on the training"""
        if not (self.check_has_trained and keep_training):
            # reset the model weights
            self._reset_model()
            self.optimizer = optimizer(self.model.parameters(), **optimizers_param)
            if lr_scheduler is not None:
                if scheduler_params:
                    self.scheduler = lr_scheduler(self.optimizer, **scheduler_params)
                else:
                    self.scheduler = lr_scheduler(self.optimizer)
        elif cross_validation:
            # reset the model weights
            self._reset_model()
            # do not re-initialise the optimizer if keep-training
            dict_param = self.optimizer.param_groups[0]
            dict_param.pop("params", None)  # model.parameters()
            dict_param.pop("initial_lr", None)
            self.optimizer.__init__(self.model.parameters(), **dict_param)  # type: ignore
            if lr_scheduler is not None:  # reset scheduler
                if scheduler_params:
                    self.scheduler = lr_scheduler(self.optimizer, **scheduler_params)
                else:
                    self.scheduler = lr_scheduler(self.optimizer)

    def train(
        self,
        optimizer: Type[Optimizer],
        n_epochs: int = 10,
        cross_validation: bool = False,
        optimizers_param: Optional[Dict[str, Any]] = None,
        dataloaders_param: Optional[Dict[str, Any]] = None,
        lr_scheduler: Optional[Type[_LRScheduler]] = None,
        scheduler_params: Optional[Dict[str, Any]] = None,
        optuna_params: Optional[Tuple[BaseTrial, str]] = None,
        profiling: bool = False,
        parallel_tpu: bool = False,
        keep_training: bool = False,
        store_grad_layer_hist: bool = False,
        n_accumulated_grads: int = 0,
        writer_tag: str = "",
    ) -> Tuple[float, float]:
        """Function to run all the training cycles.

        Args:
            optimizer:
                the torch optimiser class, like `SGD`
            n_epochs :
                number of training epochs
            cross_validation:
                whether or not to perform five-fold cross-validation
            dataloaders_param:
                dictionary of the dataloaders
                parameters, e.g. `{'batch_size': 32}`. If ``None``, then
                the parameters of the training and validation
                dataloaders will be used.
            optimizers_param:
                dictionary of the optimizers
                parameters, e.g. `{"lr": 0.001}`
            lr_scheduler:
                a learning rate scheduler class
            scheduler_params:
                learning rate scheduler parameters
            optuna_params :
                the parameters `(trial, search_metric)`
                used in the gridsearch. Saefly ignore for
                standard trainings
            profiling:
                whether or not you want to activate the
                profiler
            parallel_tpu:
                Use or not parallel TPU cores.
                Still experimental!
            keep_training:
                This flag allows to restart a training from
                the existing optimizer as well as the
                existing model
            store_grad_layer_hist:
                This flag allows to store the gradients
                and the layer values in tensorboard for
                each epoch
            n_accumulated_grads:
                this is the number of accumulated gradients.
                Only a positive number will be taken into account
            writer_tag:
                the tensorboard writer tag

        Returns:
            (float, float):
                the validation loss and accuracy
                if there is cross validation, the validation data loader
                is ignored. On the other hand, if there `cross_validation = False`
                then the test loss and accuracy is returned.
        """
        self.n_accumulated_grads = n_accumulated_grads
        self.store_grad_layer_hist = store_grad_layer_hist
        # to start the training from where we left
        if self.check_has_trained and keep_training:
            self._set_initial_model()

        # train initialisation
        dl_tr = self.dataloaders[0]
        if optimizers_param is None:
            optimizers_param = {"lr": 0.001}

        # dataloaders_param initialisation
        if dataloaders_param is None:
            if self.dataloaders[1] is not None:
                dataloaders_param_val = Trainer.copy_dataloader_params(
                    self.dataloaders[1]
                )
            else:
                dataloaders_param_val = Trainer.copy_dataloader_params(dl_tr)
            dataloaders_param_tr = Trainer.copy_dataloader_params(dl_tr)
        else:
            dataloaders_param_val = dataloaders_param.copy()
            dataloaders_param_tr = dataloaders_param.copy()

        # scheduler_params initialisation
        if scheduler_params is None:
            scheduler_params = {}

        # LR scheduler
        if not (self.check_has_trained and keep_training):
            self.scheduler = None

        # optuna gridsearch
        search_metric = None
        trial = None
        if optuna_params is not None:
            check_optuna = True
            trial, search_metric = optuna_params
        else:
            check_optuna = False

        # profiling
        self._init_profiler(
            profiling, cross_validation, n_epochs, self.k_fold_class.n_splits  # type: ignore
        )

        # remove sampler to avoid conflicts with indexing
        # we will re-introduce the sampler when creating the indexing list
        try:
            dataloaders_param_val.pop("sampler")
        except KeyError:
            pass

        try:
            dataloaders_param_tr.pop("sampler")
        except KeyError:
            pass

        # validation being the 20% in the case of 2
        # dataloders without crossvalidation
        if len(self.dataloaders) == 3:  # type: ignore
            try:
                val_idx = self.dataloaders[1].sampler.indices  # type: ignore
            except AttributeError:
                val_idx = list(range(len(self.dataloaders[1].dataset)))  # type: ignore
            # print(val_idx)
            dl_val = torch.utils.data.DataLoader(  # type: ignore
                self.dataloaders[1].dataset,
                # pin_memory=True,
                **dataloaders_param_val,
                sampler=SubsetRandomSampler(val_idx),
            )
            try:
                tr_idx = self.dataloaders[0].sampler.indices  # type: ignore
            except AttributeError:
                tr_idx = list(range(len(self.dataloaders[0].dataset)))  # type: ignore
            # print(tr_idx)
            dl_tr = torch.utils.data.DataLoader(  # type: ignore
                self.dataloaders[0].dataset,
                # pin_memory=True,
                **dataloaders_param_tr,
                sampler=SubsetRandomSampler(tr_idx),
            )
        else:
            try:
                data_idx = self.dataloaders[0].sampler.indices  # type: ignore
            except AttributeError:
                data_idx = list(range(len(self.dataloaders[0].dataset)))  # type: ignore
            # print(data_idx)
            tr_idx, val_idx = train_test_split(data_idx, test_size=0.2)
            dl_val = torch.utils.data.DataLoader(  # type: ignore
                self.dataloaders[0].dataset,
                # pin_memory=True,
                **dataloaders_param_val,
                sampler=SubsetRandomSampler(val_idx),
            )
            dl_tr = torch.utils.data.DataLoader(  # type: ignore
                self.dataloaders[0].dataset,
                # pin_memory=True,
                **dataloaders_param_tr,
                sampler=SubsetRandomSampler(tr_idx),
            )

        if cross_validation:
            mean_val_loss = []
            mean_val_acc = []
            try:
                data_idx = self.dataloaders[0].sampler.indices  # type: ignore
                labels_for_split = [
                    self.dataloaders[0].dataset[i][-1] for i in data_idx
                ]
            except AttributeError:
                data_idx = list(range(len(self.dataloaders[0].dataset)))  # type: ignore
                labels_for_split = [
                    self.dataloaders[0].dataset[i][-1] for i in data_idx
                ]

            for fold, (tr_idx, val_idx) in enumerate(
                self.k_fold_class.split(data_idx, labels_for_split)
            ):
                # prints for class balance
                # lab_tr_fold = [self.dataloaders[0].dataset[i][-1] for i in tr_idx]
                # lab_val_fold = [self.dataloaders[0].dataset[i][-1] for i in val_idx]
                # print("train labels:",[(i, lab_tr_fold.count(i)) for i in np.unique(np.array(lab_tr_fold))])
                # print("val labels:",[(i, lab_val_fold.count(i)) for i in np.unique(np.array(lab_val_fold))])
                # print("lenghts: ", len(lab_tr_fold), len(lab_val_fold))
                self._init_optimizer_and_scheduler(
                    keep_training,
                    cross_validation,
                    optimizer,
                    optimizers_param,
                    lr_scheduler,
                    scheduler_params,
                )

                # re-initialise data loaders
                if len(self.dataloaders) == 3:
                    warnings.warn(
                        "Validation set is ignored in automatic Cross Validation"
                    )
                dl_tr = torch.utils.data.DataLoader(  # type: ignore
                    self.dataloaders[0].dataset,
                    # pin_memory=True,
                    **dataloaders_param_tr,
                    sampler=SubsetRandomSampler(tr_idx),
                )
                dl_val = torch.utils.data.DataLoader(  # type: ignore
                    self.dataloaders[0].dataset,
                    # pin_memory=True,
                    **dataloaders_param_val,
                    sampler=SubsetRandomSampler(val_idx),
                )
                # print n-th fold
                print("\n\n********** Fold ", fold + 1, "**************")
                # the training and validation loop
                if parallel_tpu == False:
                    valloss, valacc = self._training_loops(
                        n_epochs,
                        dl_tr,
                        dl_val,
                        lr_scheduler,
                        self.scheduler,
                        check_optuna,
                        search_metric,
                        trial,
                        cross_validation,
                        writer_tag + "/fold = " + str(fold + 1),
                    )
                else:
                    valloss, valacc = self.parallel_tpu_training_loops(
                        n_epochs,
                        dl_tr,
                        dl_val,
                        optimizers_param,
                        lr_scheduler,
                        self.scheduler,
                        check_optuna,
                        search_metric,
                        trial,
                        cross_validation,
                    )

                mean_val_loss.append(valloss)
                mean_val_acc.append(valacc)
            # mean of the validation and loss accuracies over folds
            if self.best_not_last:
                valloss = min(
                    np.array(
                        _inner_refactor_scalars(
                            self.val_loss_list_hparam, True, self.k_fold_class.n_splits  # type: ignore
                        )
                    )[:, 0]
                )
                valacc = max(
                    np.array(
                        _inner_refactor_scalars(
                            self.val_acc_list_hparam, True, self.k_fold_class.n_splits  # type: ignore
                        )
                    )[:, 0]
                )
            else:
                valloss = torch.mean(torch.tensor(mean_val_loss)).item()
                valacc = torch.mean(torch.tensor(mean_val_acc)).item()

        else:
            self._init_optimizer_and_scheduler(
                keep_training,
                cross_validation,
                optimizer,
                optimizers_param,
                lr_scheduler,
                scheduler_params,
            )

            if not parallel_tpu:
                valloss, valacc = self._training_loops(
                    n_epochs,
                    dl_tr,
                    dl_val,
                    lr_scheduler,
                    self.scheduler,
                    check_optuna,
                    search_metric,
                    trial,
                    False,
                    writer_tag,
                )
            else:
                valloss, valacc = self.parallel_tpu_training_loops(
                    n_epochs,
                    dl_tr,
                    dl_val,
                    optimizers_param,
                    lr_scheduler,
                    self.scheduler,
                    check_optuna,
                    search_metric,
                    trial,
                    False,
                )

        try:
            self.writer.flush()  # type: ignore
        except AttributeError:
            pass
        # check for training
        self.check_has_trained = True

        # put the mean of the cross_val
        return valloss, valacc

    def _training_loops(
        self,
        n_epochs: int,
        dl_tr: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        dl_val: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        lr_scheduler: Optional[Type[_LRScheduler]] = None,
        scheduler: Optional[_LRScheduler] = None,
        check_optuna: bool = False,
        search_metric: Optional[str] = None,
        trial: Optional[BaseTrial] = None,
        cross_validation: bool = False,
        writer_tag: str = "",
    ) -> Tuple[float, float]:
        """private method to run the trainign loops

        Args:
            n_epochs:
                number of training epochs
            dl_tr:
                training dataloader
            dl_val:
                validation dataloader
                parameters, e.g. `{'batch_size': 32}`
            lr_scheduler:
                a learning rate scheduler class
            scheduler:
                the actual scheduler instance
            check_optuna:
                boolean to store the optuna results of
                the trial
            search_metric:
                either ``'loss'`` or ``'accuracy'``, this
                corresponds to the gridsearch criterion
            trial:
                the optuna trial
            cross_validation (bool, default False)
                the boolean flag for cross validation
            writer_tag:
                the tensorboard writer tag

        Returns:
            (float, float):
                the validation loss and validation accuracy
        """

        valloss, valacc = np.inf, 0.0
        for t in range(n_epochs):
            print(f"Epoch {t + 1}\n-------------------------------")
            self.val_epoch = t
            self.train_epoch = t
            self._train_loop(dl_tr, writer_tag)
            me = ModelExtractor(self.model, self.loss_fn)
            if self.store_grad_layer_hist:
                try:
                    lista = me.get_layers_param()
                    for k, item in lista.items():
                        try:
                            self.writer.add_histogram(  # type: ignore
                                writer_tag + "/weights&biases/param/train/" + k, item, t
                            )
                            self.writer.add_histogram(  # type: ignore
                                writer_tag + "/weights&biases/param/train/log/" + k,
                                torch.log(torch.abs(item) + 1e-8),
                                t,
                            )
                        except ValueError:
                            warnings.warn(
                                f"The histogram is empty, most likely because your loss"
                                f" is exploding. Try use gradient clipping."
                            )
                    lista_grad = me.get_layers_grads()
                    for k, item in zip(lista.keys(), lista_grad):
                        try:
                            self.writer.add_histogram(  # type: ignore
                                writer_tag + "/weights&biases/grads/train/" + k, item, t
                            )
                            self.writer.add_histogram(  # type: ignore
                                writer_tag + "/weights&biases/param/train/log/" + k,
                                torch.log(torch.abs(item) + 1e-8),
                                t,
                            )
                        except ValueError:
                            warnings.warn(
                                f"The histogram is empty, most likely because your loss"
                                f" is exploding. Try use gradient clipping."
                            )
                    self.writer.flush()  # type: ignore

                except AttributeError:
                    pass
            self._run_pipe_hook(t + 1, self.optimizer, me, self.writer)
            valloss, valacc = self._val_loop(dl_val, writer_tag)
            # absolute best of loss and accuracy
            self.best_val_acc = max(self.best_val_acc, valacc)
            self.best_val_loss = min(self.best_val_loss, valloss)
            # print(self.optimizer.param_groups[0]["lr"])
            if not lr_scheduler is None:
                scheduler.step()  # type: ignore
            # pruning trials
            if check_optuna and not cross_validation:
                if search_metric == "loss":
                    trial.report(valloss, t)  # type: ignore
                else:
                    trial.report(valacc, t)  # type: ignore
                # Handle pruning based on the intermediate value.
                if trial.should_prune():  # type: ignore
                    raise optuna.exceptions.TrialPruned()
        return valloss, valacc

    def parallel_tpu_training_loops(
        self,
        n_epochs: int,
        dl_tr_old: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        dl_val_old: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]],
        optimizers_param: Dict[str, Any],
        lr_scheduler: Optional[Type[_LRScheduler]] = None,
        scheduler: Optional[_LRScheduler] = None,
        check_optuna: bool = False,
        search_metric: Optional[str] = None,
        trial: Optional[BaseTrial] = None,
        cross_validation: bool = False,
    ) -> Tuple[float, float]:
        """Experimental function to run all the training cycles
        on colab TPUs in parallel.
        Note: ``cross_validation`` parameter as well as
        ``profiling`` are ignored.

        Args:
            n_epochs:
                number of training epochs
            dl_tr_old:
                training dataloader
            dl_val_old:
                validation dataloader
                parameters, e.g. `{'batch_size': 32}`
            optimizers_param:
                dictionary of parameters for the optimizers
            lr_scheduler :
                a learning rate scheduler class
            scheduler:
                the actual scheduler instance
            check_optuna :
                boolean to store the optuna results of
                the trial
            search_metric:
                either ``'loss'`` or ``'accuracy'``, this
                corresponds to the gridsearch criterion
            trial:
                the optuna trial
            cross_validation :
                the boolean flag for cross validation

        Returns:
            (float, float):
                the validation loss and validation accuracy
        """
        self.val_loss = 0
        self.val_acc = 0
        warnings.warn(
            "The tensorboard writer is ignored "
            + "for multi TPU processing. Also SAM optimisation"
            + " does not work for multi TPU training."
        )

        def map_fun_custom(index, flags):
            """map function for multi-processing"""
            device = xm.xla_device()

            print("uploading model to TPU")
            model2 = self.model.to(device)

            # initialize optimizer
            optimizer_class = type(self.optimizer)
            optimizer = optimizer_class(model2.parameters(), **optimizers_param)

            # define training and validation
            # distributed samplers and update
            # the dataloaders
            train_sampler = torch.utils.data.distributed.DistributedSampler(  # type: ignore
                dl_tr_old.dataset,
                num_replicas=xm.xrt_world_size(),
                rank=xm.get_ordinal(),
                shuffle=True,
            )

            dl_tr = torch.utils.data.DataLoader(  # type: ignore
                dl_tr_old.dataset,
                num_workers=dl_tr_old.num_workers,
                batch_size=dl_tr_old.batch_size,
                sampler=train_sampler,
                drop_last=dl_tr_old.drop_last,
            )

            val_sampler = torch.utils.data.distributed.DistributedSampler(  # type: ignore
                dl_val_old.dataset,
                num_replicas=xm.xrt_world_size(),
                rank=xm.get_ordinal(),
                shuffle=False,
            )

            dl_val = torch.utils.data.DataLoader(  # type: ignore
                dl_val_old.dataset,
                num_workers=dl_val_old.num_workers,
                batch_size=dl_val_old.batch_size,
                sampler=val_sampler,
                drop_last=dl_val_old.drop_last,
            )

            # train loop
            for t in range(n_epochs):
                model2.train()
                para_train_loader = pl.ParallelLoader(
                    dl_tr, [device]
                ).per_device_loader(device)
                print(f"Epoch {t + 1}\n-------------------------------")
                self.val_epoch = t
                self.train_epoch = t

                # train batch loop
                loss = 0.0
                correct = 0.0
                tik = time.time()
                # for batch, (X, y) in enumerate(self.dataloaders[0]):
                for batch, (X, y) in enumerate(para_train_loader):
                    # Compute prediction and loss
                    pred = model2(X)
                    try:
                        correct += (
                            (pred.argmax(1) == y).to(torch.float).sum().item()
                        )  # noqa
                    except RuntimeError:
                        correct += (
                            (pred.argmax(2) == y).to(torch.float).sum().item()
                        )  # noqa
                    loss = self.loss_fn(pred, y)
                    # Save to tensorboard
                    # self.writer.add_scalar("Parallel" + "/Loss/train",
                    #                       loss.cpu(),
                    #                       self.train_epoch*dl_tr.batch_size + batch)
                    # Backpropagation
                    self.optimizer.zero_grad()
                    loss.backward()

                    xm.optimizer_step(self.optimizer)

                # train accuracy:
                correct /= len(dl_tr) * dl_tr.batch_size
                print("Train accuracy at epoch ", t, " : ", correct)
                print(f"\nTime taken for this epoch: {round(time.time() - tik):.2f}s")

                # evaluation
                model2.eval()

                loss, correct = 0.0, 0.0
                class_label: List[Tensor] = []
                class_probs: List[List[Tensor]] = []

                pred = 0.0
                para_valid_loader = pl.ParallelLoader(
                    dl_val, [device]
                ).per_device_loader(device)
                with torch.no_grad():
                    # per batch!!
                    for X, y in para_valid_loader:
                        pred = model2(X)
                        class_probs_batch = [f.softmax(el, dim=0) for el in pred]
                        class_probs.append(class_probs_batch)
                        loss += self.loss_fn(pred, y).item()
                        try:
                            correct += (
                                (pred.argmax(1) == y).to(torch.float).sum().item()
                            )
                        except RuntimeError:
                            correct += (
                                (pred.argmax(2) == y).to(torch.float).sum().item()
                            )
                        class_label.append(y)
                    # add data to tensorboard
                    # self._add_pr_curve_tb(pred, class_label, class_probs, "validation")

                # self.writer.add_scalar("Parallel " + "/Accuracy/validation", correct, self.val_epoch)
                print(
                    f"Validation results: \n Accuracy: {(100 * correct):.2f}%, \
                        Avg loss: {loss:>8f} \n"
                )

                # self.writer.flush()

                if lr_scheduler is not None:
                    assert scheduler is not None, "scheduler is None"
                    scheduler.step()
                if self.prof is not None:
                    self.prof.step()

                if check_optuna and not cross_validation:
                    if search_metric == "loss":
                        assert trial is not None, "trial is None"
                        trial.report(loss, t)
                    else:
                        assert trial is not None, "trial is None"
                        trial.report(correct, t)
                    # Handle pruning based on the intermediate value.
                    if trial.should_prune():
                        raise optuna.exceptions.TrialPruned()
            self.val_loss += loss  # type: ignore
            self.val_acc += correct * 100  # type: ignore

        flags: Dict[Any, Any] = {}
        self.val_acc /= len(dl_val_old) * dl_val_old.batch_size  # type: ignore
        self.val_loss /= len(dl_val_old)  # type: ignore
        xmp.spawn(map_fun_custom, args=(flags,), nprocs=8, start_method="fork")
        return self.val_loss, self.val_acc

    def evaluate_classification(
        self,
        num_class: Optional[int] = None,
        dl: Optional[DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]]] = None,
    ) -> Tuple[float, float, np.ndarray]:
        """Method to evaluate the performance of the model.

        Args:
            num_class:
                number of classes
            dl :
                the Dataloader to evaluate. If ``None``,
                we use the training dataloader in ``self``

        Returns:
            (float, float, 2darray):
                the accuracy, loss and confusion matrix.
        """
        if dl is None:
            dl = self.dataloaders[0]
        class_probs: List[List[Tensor]] = []
        class_label: List[Tensor] = []
        batch_metric_list = []
        loss = 0.0
        correct = 0.0
        confusion_matrix = np.zeros((num_class, num_class))  # type: ignore
        self.model.eval()
        with torch.no_grad():
            for batch, (X, y) in tqdm(enumerate(dl)):
                pred, X, y = self._send_to_device(X, y)
                class_probs_batch = [f.softmax(el, dim=0) for el in pred]
                class_probs.append(class_probs_batch)
                loss += self.loss_fn(pred, y).item()
                batch_metric = self.training_metric(pred, y)
                batch_metric_list.append(batch_metric)
                class_label.append(y)
                for t, p in zip(y.view(-1), pred.argmax(1).view(-1)):
                    confusion_matrix[t.long(), p.long()] += 1
        epoch_metric = sum(batch_metric_list) / len(batch_metric_list)
        epoch_loss = loss / len(batch_metric_list)
        return epoch_metric, epoch_loss, confusion_matrix

    def register_pipe_hook(
        self,
        callable: Callable[
            [int, Optimizer, ModelExtractor, Optional[SummaryWriter]], Any
        ],
    ) -> None:
        """This method registers a function that
        will be called after each trainign step.

        The arguments of the callable function are, in this order:
         - current epoch number
         - current optimizer instance
         - the ModelExtractor instance at that epoch
         - the tensorboard writer

        Args:
            callable (Callable):
                the function to register"""
        self.registered_hook = callable

    def _run_pipe_hook(
        self,
        epoch: int,
        optim: Optimizer,
        me: ModelExtractor,
        writer: Optional[SummaryWriter] = None,
    ) -> None:
        """private method that runs the hooked
        function at every epoch, after the single training loop"""
        if self.registered_hook is not None:
            self.registered_hook(epoch, optim, me, writer)

    @staticmethod
    def copy_dataloader_params(
        original_dataloader: DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]]
    ) -> Dict[str, Any]:
        """returns the dict of init parameters"""
        return {
            "batch_size": original_dataloader.batch_size,
            # "batch_sampler": original_dataloader.batch_sampler,
            "num_workers": original_dataloader.num_workers,
            "collate_fn": original_dataloader.collate_fn,
            "pin_memory": original_dataloader.pin_memory,
            "drop_last": original_dataloader.drop_last,
            "timeout": original_dataloader.timeout,
            "worker_init_fn": original_dataloader.worker_init_fn,
            "multiprocessing_context": original_dataloader.multiprocessing_context,
            "generator": original_dataloader.generator,
            "prefetch_factor": original_dataloader.prefetch_factor,
            "persistent_workers": original_dataloader.persistent_workers,
        }



import os
import shutil
from typing import Any, Callable, Dict, Optional, Optional, Tuple, TypeVar, Union

import numpy as np
import torch

from sympy import false
from torch.utils.data import DataLoader, Dataset

from gdeep.utility.custom_types import Array
from gdeep.utility.custom_types import Tensor


class FromArray(Dataset[Tuple[Tensor, Tensor]]):
    """This class is useful to build dataloaders
    from a array of X and y. Tensors are also supported.

    Args:
        X :
            The data. The first dimension is the datum index
        y :
            The labels, need to match the first dimension
            with the data

    """

    def __init__(self, x: Union[Tensor, Array], y: Union[Tensor, Array]) -> None:
        self.x = self._from_numpy(x)
        y = self._from_numpy(y)
        self.y = self._long_or_float(y)

    @staticmethod
    def _from_numpy(x: Union[Tensor, Array]) -> Tensor:
        """this is an upgrade of ``torch.from_numpy()``
        that also allows tensor input.
        """
        if isinstance(x, torch.Tensor):
            return x
        return torch.from_numpy(x)

    @staticmethod
    def _long_or_float(y: Union[Tensor, Array]) -> Tensor:
        """This private method converts the labels to either
        a long tensor of a float tensor"""
        if isinstance(y, torch.Tensor):
            return y
        if y.dtype in [np.float16, np.float32, np.float64]:  # type: ignore
            return torch.tensor(y).float()
        return torch.tensor(y).long()

    def __len__(self):
        return self.y.shape[0]

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        x, y = (self.x[idx], self.y[idx])
        return x, y



from abc import ABC, abstractmethod
from typing import Optional

import torch
from gdeep.topology_layers.persformer_config import PersformerConfig
from torch.nn import Module

# Type aliases
from gdeep.utility.custom_types import Tensor


class AttentionBase(Module, ABC):
    """Base class for attention layers. This class
    can be used in generic transformer models.
    """

    def __init__(self, config: PersformerConfig):
        super().__init__()
        self.config = config

    @abstractmethod
    def forward(
        self, input: Tensor, attention_mask: Optional[Tensor] = None  # type: ignore
    ) -> Tensor:
        raise NotImplementedError



import os

import numpy as np
from gdeep.data.datasets.persistence_diagrams_from_graphs_builder import (
    PersistenceDiagramFromGraphBuilder,
)
from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
    get_one_hot_encoded_persistence_diagram_from_gudhi_extended,
)
from gdeep.utility import DEFAULT_GRAPH_DIR


def test_mutag_load_and_save():

    # Parameters
    name_graph_dataset: str = "MUTAG"
    diffusion_parameter: float = 0.1
    num_homology_types: int = 4

    # Create the persistence diagram dataset
    pd_creator = PersistenceDiagramFromGraphBuilder(
        name_graph_dataset, diffusion_parameter
    )
    pd_creator.create()

    gudhi_ext = (
        np.array(
            [
                [0.90634646, 0.90835863],
                [0.90659781, 0.90835863],
                [0.90672356, 0.90684934],
                [0.90672356, 0.90684934],
            ]
        ),
        np.array([[0.90634646, 0.90835863]]),
        np.array(
            [
                [0.90659795, 0.90710085],
                [0.90672356, 0.90710077],
                [0.90672356, 0.90684934],
            ]
        ),
        np.array([[0.90659781, 0.90710077], [0.90659781, 0.90710085]]),
    )

    pd_gudhi = get_one_hot_encoded_persistence_diagram_from_gudhi_extended(gudhi_ext)

    # Print sample extended persistence diagram
    file_path: str = os.path.join(
        DEFAULT_GRAPH_DIR,
        f"MUTAG_{diffusion_parameter}_extended_persistence",
        "diagrams",
    )
    graph_idx = 1
    pd = OneHotEncodedPersistenceDiagram.load(
        os.path.join(file_path, f"{graph_idx}.npy")
    )
    assert pd_gudhi.all_close(
        pd, atol=1e-2
    ), "Generated persistence diagram is not equal to the one from GUDHI"



from typing import List, Any, Optional

import numpy
import torch
from sklearn.metrics import pairwise_distances
from gtda.homology import VietorisRipsPersistence, WeakAlphaPersistence
from gtda.graphs import KNeighborsGraph, GraphGeodesicDistance


from gdeep.utility.custom_types import Tensor, Array


def knn_distance_matrix(x: List, k: int = 3):

    """Returns list of distance matrices according
    to the k-NN graph distance of the datasets X

    Args:
        x:
            the point cloud
        k (int):
            the numbers of neighbors for the k-NN graph

    """

    kng = KNeighborsGraph(n_neighbors=k)
    adj = kng.fit_transform(x)
    return GraphGeodesicDistance(directed=False).fit_transform(adj)


def persistence_diagrams_of_activations(
    activations_list: List[Tensor],
    homology_dimensions: Optional[List[int]] = None,
    k: int = 5,
    mode: str = "VR",
    max_edge_length: int = 10,
) -> List[Any]:
    """Returns list of persistence diagrams of the activations of all
    layers of type layer_types

    Args:
        activations_list:
            list of activation
            tensors for each layer
        homology_dimensions :
            list of homology
            dimensions. Defaults to `[0, 1]`.
        k  :
            number of neighbors parameter
            of the k-NN distance. If ``k <= 0``, then
            the list of activations is considered
            as a point cloud and no knn distance is
            computed
        mode  :
            choose the filtration ('VR'
            or 'alpha') to compute persistence default to 'VR'.
        max_edge_length :
            maximum edge length of the simplices forming
            the complex

    Returns:
        (list):
            list of persistence diagrams of activations
            of the different layers
    """
    if homology_dimensions is not None:
        homology_dimensions = [0, 1]
    for i, activ in enumerate(activations_list):
        if len(activ.shape) > 2:  # in case of non FF layers
            activations_list[i] = activ.view(activ.shape[0], -1)

    if k > 0 and mode == "VR":
        vr = VietorisRipsPersistence(
            homology_dimensions=homology_dimensions,
            metric="precomputed",
            reduced_homology=False,
            infinity_values=20,
            max_edge_length=max_edge_length,
            collapse_edges=True,
        )
        # infinity_values set to avoid troubles with
        # multiple topological features living indefinitely
    elif k <= 0 and mode == "alpha":
        vr = WeakAlphaPersistence(
            homology_dimensions=homology_dimensions,
            reduced_homology=False,
            max_edge_length=max_edge_length,
        )
    else:
        vr = VietorisRipsPersistence(
            homology_dimensions=homology_dimensions,
            reduced_homology=False,
            max_edge_length=max_edge_length,
            collapse_edges=True,
        )

    if k > 0 and mode == "VR":
        activations_list_array = _convert_list_of_tensor_to_numpy(activations_list)
        dist_matrix = knn_distance_matrix(activations_list_array, k=k)
        persistence_diagrams = vr.fit_transform(dist_matrix)
    else:
        activations_list_array = _convert_list_of_tensor_to_numpy(activations_list)
        persistence_diagrams = vr.fit_transform(activations_list_array)

    return persistence_diagrams


def _simplified_persistence_of_activations(
    inputs: List[Tensor],
    homology_dimensions: List[int],
    filtration_value: float,
    **kwargs
):
    """this method filters out the distances between the input points that are below the
    filtration value and sets to 1 those that are above it. This is useful for a simplified
    computation of homology"""
    vr = VietorisRipsPersistence(
        homology_dimensions=homology_dimensions, metric="precomputed", **kwargs
    )

    activations_list_array = _convert_list_of_tensor_to_numpy(inputs)
    activations_distances = [
        pairwise_distances(x.reshape((x.shape[0], -1))) for x in activations_list_array  # type: ignore
    ]
    activations_filtered_distances = [
        (array > filtration_value) * 1.0 for array in activations_distances
    ]

    simplified_persistence_diagrams = vr.fit_transform(activations_filtered_distances)
    return simplified_persistence_diagrams


def _convert_list_of_tensor_to_numpy(input_list: List[Tensor]) -> List[Array]:
    """private method to convert a list of tensors to a
    list of arrays"""
    output_list: List[Array] = []
    for item in input_list:
        output_list.append(item.detach().cpu().numpy())
    return output_list



from typing import List, Dict, Callable, Tuple, Union
from copy import copy

import torch

from ..analysis.decision_boundary import GradientFlowDecisionBoundaryCalculator
from ..analysis.decision_boundary import UniformlySampledPoint
from . import SaveLayerOutput
from gdeep.utility import DEVICE

from gdeep.utility.custom_types import Tensor


class ModelExtractor:
    """This class wraps nn.Modules to extract
    weights, activations, gradients and decision boundaries

    Args:
        model:
            standard torch module
        loss_fn :
            loss function

    Examples::

        from gdeep.analysis import Extractor
        # the candidate datum for a question-answering example
        x = next(iter(transformed_textts))[0].reshape(1, 2, -1).to(DEVICE)
        # the model extractor: you need a trainer and the loss function
        ex = ModelExtractor(trainer.model, loss_fn)
        # getting the names of the layers
        layer_names = ex.get_layers_param().keys()
        print("Let's extract the activations of the first attention layer: ", next(iter(layer_names)))
        self_attention = ex.get_activations(x)[:2]

    """

    def __init__(
        self, model: torch.nn.Module, loss_fn: Callable[[Tensor, Tensor], Tensor]
    ) -> None:
        # self.model = model
        self.model = model.to(DEVICE)
        self.loss_fn = loss_fn

    def _send_to_device(
        self, x: Union[Tensor, List[Tensor]], y: Tensor
    ) -> Tuple[Tensor, Union[Tensor, List[Tensor]], Tensor]:
        """use this private method to send the
        ``x`` and ``y`` to the ``DEVICE``.

        Args:
            x:
                the input of the model, either a List[Tensor] or a Tensor
            y:
                the label

        Returns:
            (Tensor, Union[Tensor, List[Tensor]], Tensor)
                the prediction for x, x and the label

        """
        new_x: List[Tensor] = []
        if isinstance(x, tuple) or isinstance(x, list):
            for xi in x:
                try:
                    xi.requires_grad = True
                except RuntimeError:  # only float and complex tensors can have grads
                    pass
                new_x.append(xi.to(DEVICE))
            x = copy(new_x)

        else:
            try:
                x.requires_grad = True
            except RuntimeError:  # only float and complex tensors can have grads
                pass
            x = x.to(DEVICE)

        y = y.to(DEVICE)
        # Compute prediction and loss
        if isinstance(x, tuple) or isinstance(x, list):
            prediction = self.model(*x)
        else:
            prediction = self.model(x)
        return prediction, x, y

    def get_decision_boundary(
        self, input_example: Tensor, n_epochs: int = 100, precision: float = 0.1
    ) -> Tensor:
        """Compute the decision boundary for self.model
        with self.loss_fn

        Args:
            n_epochs:
                number of training cycles to find
                the decision boundary
            input_example:
                an example of a single input,
                to extract the dimensions of the feature space
            precision:
                parameter to filter the spurious data that are
                close to te decision boundary, but not close enough

        Returns:
            Tensor:
                the pointcloud defining the decision
                boundary with dimensions `(n_samples, n_features)`
        """
        input_dim = input_example.flatten().shape[0]
        bounding_box = [(-1, 1) for _ in range(input_dim)]
        us = UniformlySampledPoint(bounding_box, n_samples=3000)
        sample_points_tensor = torch.from_numpy(us()).to(torch.float)
        # reshape points as example
        sample_points_tensor = sample_points_tensor.reshape(
            -1, *input_example.shape
        ).to(DEVICE)

        sample_points_tensor = sample_points_tensor.to(DEVICE)
        # print(sample_points_tensor.shape)
        # Using new gradient flow implementation
        self.model.train()
        gf = GradientFlowDecisionBoundaryCalculator(
            model=self.model,
            initial_points=sample_points_tensor,
            optimizer=lambda params: torch.optim.Adam(params),
        )
        gf.step(number_of_steps=n_epochs)
        res = gf.get_filtered_decision_boundary(delta=precision).detach()
        return res

    def get_activations(self, x: Union[Tensor, List[Tensor]]) -> List[Tensor]:
        """Compute the activations of self.model with input
        `X`

        Args:
            x :
                an example of an input or
                an input batch of which to compute the activation(s)

        Returns:
            list:
                list of the activation Tensors
        """

        saved_output_layers = SaveLayerOutput()

        hook_handles = []
        for layer in self.model.modules():
            handle = layer.register_forward_hook(saved_output_layers)
            hook_handles.append(handle)
        self.model.eval()
        _, x, _ = self._send_to_device(
            x, torch.rand((1, 1))
        )  # here we also run the model(x)

        for handle in hook_handles:
            handle.remove()

        self.model.train()
        return saved_output_layers.get_outputs()

    def get_layers_param(self) -> Dict[str, Tensor]:
        """Returns parameters of layers

        Returns:
            dict:
                dict of tensors, corresponding
                to the layer parameters. The key of
                the dict is the name of the parameters
        """

        # layer_data = dict()
        # for name, layer_param in self.model.named_parameters():
        #    layer_data[name]=layer_param.data
        return self.model.state_dict()

    def get_layers_grads(self) -> List[Tensor]:
        """Returns the gradients of each layer

        Returns:
            list:
                list of tensors, corresponding
                to the layer gradients (weights and biases!).
        """

        # layer_data = dict()
        # for name, layer_param in self.model.named_parameters():
        #    layer_data[name]=layer_param.data
        sdict = self.model.parameters()
        output = []
        for v in sdict:
            output.append(v.grad)
        return output

    def get_gradients(
        self, batch: Tuple[Union[Tensor, List[Tensor]], Tensor], **kwargs
    ) -> Tuple[List[Tensor], List[Tensor]]:
        """Returns the **averaged gradient** of the self.loss_fn.
        To specify the target variable (e.g. the true class),
        use the keywords argument `target=`

        Args:
            batch:
                a data batch on which to compute the average
                gradients. f the batch has one single item, then
                it will output the gradients for that single datum

        Returns:
            list,  list:
                the gradients for the inputs; the list of tensors,
                corresponding to the gradient of the weights.
        """
        x, target = batch
        pred, x, target = self._send_to_device(x, target)
        loss = self.loss_fn(pred, target, **kwargs)

        for k, param in self.model.state_dict().items():
            if param.dtype is torch.float:
                param.requires_grad = True
                param.retain_grad()  # this hook allows to store grad
        loss.backward()
        output_grads = []
        for param in self.model.parameters():
            if param.dtype is torch.float:
                output_grads.append(param.grad)
        grads: List[Tensor] = []
        if isinstance(x, tuple) or isinstance(x, list):
            for xi in x:
                grads.append(xi.grad)  # type: ignore
        else:
            grads.append(x.grad)  # type: ignore
        return grads, output_grads



# %%
from os import remove, makedirs, environ
from os.path import join, exists
import logging

import google  # type: ignore
from google.cloud import storage  # type: ignore
from google.cloud.storage import Bucket  # type: ignore
from google.auth.exceptions import DefaultCredentialsError  # type: ignore
import hashlib
import pytest
import random
from shutil import rmtree

from gdeep.data.datasets import _DataCloud
from gdeep.utility.utils import get_checksum
from gdeep.utility import DATASET_BUCKET_NAME

LOGGER = logging.getLogger(__name__)


if "GOOGLE_APPLICATION_CREDENTIALS" in dict(environ):
    # Check if the credentials are valid and if the bucket can be accessed
    client = storage.Client()
    if Bucket(client, DATASET_BUCKET_NAME).exists():

        def test_download():
            """Test download of sample data from bucket"""
            data_cloud = _DataCloud(use_public_access=False)
            file_name = "giotto-deep-big.png"
            data_cloud.download_file(file_name)

            # check if correct extension is raised when trying to download
            # non-existing file
            non_existing_file_name: str = "giotto-deep-bigs.png"
            with pytest.raises(google.api_core.exceptions.NotFound):  # type: ignore
                data_cloud.download_file(non_existing_file_name)

            # check if downloaded file exists
            file_path = join(data_cloud.download_directory, file_name)
            assert exists(file_path)

            # check if downloaded file is correct
            assert "d4b12b2dc2bc199831ba803431184fcb" == get_checksum(file_path)

            remove(join(data_cloud.download_directory, file_name))

        def test_upload():
            """Test upload of sample file to bucket."""
            data_cloud = _DataCloud(use_public_access=False)

            # create temporary file to upload to bucket
            sample_file_name = "tmp.txt"
            sample_text = "Create a new tmp file!" + str(random.randint(0, 1_000))

            if exists(sample_file_name):
                remove(sample_file_name)
            with open(sample_file_name, "w") as f:
                f.write(sample_text)

            assert exists(sample_file_name)

            # upload sample file to bucket
            data_cloud.upload_file(sample_file_name)

            # check if uploading to an already existing file raises exception
            with pytest.raises(RuntimeError):
                data_cloud.upload_file(sample_file_name)

            # delete local temporary file
            remove(sample_file_name)

            data_cloud.download_file(sample_file_name)

            data_cloud.delete_blob(sample_file_name)

            # check if downloaded file exists
            file_path = join(data_cloud.download_directory, sample_file_name)
            assert exists(file_path)

            with open(file_path, "r") as f:
                assert f.read() == sample_text

            remove(file_path)

        def test_upload_folder():
            """Test the upload of a folder to bucket and download the
            folder."""
            data_cloud = _DataCloud(use_public_access=False)

            # create temporary folder structure and temporary file to upload
            # to bucket
            # tmp: tmp.txt
            # |- sub_tmp_1: tmp1.txt
            # |- sub_tmp_2: tmp2.txt
            #     |- sub_tmp_2_2: tmp2_2.txt
            if exists("tmp"):
                rmtree("tmp")

            tmp_files = []

            sample_dir = "tmp"
            makedirs(sample_dir)
            tmp_files.append(join(sample_dir, "tmp.txt"))

            sub_1_sample_dir = join(sample_dir, "sub_tmp_1")
            makedirs(sub_1_sample_dir)
            tmp_files.append(join(sub_1_sample_dir, "tmp1.txt"))

            sub_2_sample_dir = join(sample_dir, "sub_tmp_2")
            makedirs(sub_2_sample_dir)
            tmp_files.append(join(sub_2_sample_dir, "tmp2.txt"))

            sub_2_2_sample_dir = join(sub_2_sample_dir, "sub_tmp_2_2")
            makedirs(sub_2_2_sample_dir)
            tmp_files.append(join(sub_2_2_sample_dir, "tmp2_2.txt"))

            sample_texts = {}

            for file in tmp_files:
                if exists(file):
                    remove(file)
                sample_text = "Create a new tmp file! " + str(random.randint(0, 1_000))
                sample_texts[file] = sample_text
                with open(file, "w") as f:
                    f.write(sample_text)
                assert exists(file)

            # upload sample file to bucket
            data_cloud.upload_folder(sample_dir, "tmp")

            # delete local tmp folder
            rmtree(sample_dir)

            # download folder to local
            data_cloud.download_folder(sample_dir)

            # delete folder in bucket
            data_cloud.delete_blobs("tmp")

            # check if downloaded folder is correct

            for file in sample_texts.keys():
                with open(join(data_cloud.download_directory, file), "r") as f:
                    assert f.read() == sample_texts[file]

            # delete local tmp folder
            rmtree(join(data_cloud.download_directory, sample_dir))



from collections import Counter, OrderedDict
from typing import Callable, Tuple, List, Optional, Dict, Union
from functools import partial

import torch
from torch.utils.data import Dataset
from torchtext.vocab import vocab
from torchtext.vocab.vocab import Vocab
from torchtext.data.utils import get_tokenizer  # type: ignore
from gdeep.utility import DEVICE

from ..abstract_preprocessing import AbstractPreprocessing
from .._utils import MissingVocabularyError

# type definition
from gdeep.utility.custom_types import Tensor


class TokenizerQA(
    AbstractPreprocessing[
        Tuple[str, str, List[str], List[int]],
        Tuple[Union[Tensor, List[Tensor]], Tensor],
    ]
):
    """Class to preprocess text dataloaders for Q&A
    tasks. The type of dataset is assumed to be of the
    form ``(string,string,list[string], list[string])``.

    Args:
        vocabulary:
            the torch vocabulary
        tokenizer :
            the tokenizer of the source text

    Examples::

        from gdeep.data import TorchDataLoader
        from gdeep.data import TransformingDataset
        from gdeep.data.preprocessors import TokenizerQA

        dl = TorchDataLoader(name="SQuAD2", convert_to_map_dataset=True)
        dl_tr, dl_ts = dl.build_dataloaders()

        textds = TransformingDataset(dl_tr_str.dataset,
                               TokenizerQA())

    """

    is_fitted: bool
    max_length: int
    vocabulary: Optional[Vocab]
    tokenizer: Optional[partial]
    counter: Dict[str, int]
    pad_item: int
    ordered_dict: OrderedDict

    def __init__(
        self, vocabulary: Optional[Vocab] = None, tokenizer: Optional[partial] = None
    ):
        if tokenizer is None:
            self.tokenizer = get_tokenizer("basic_english")
        else:
            self.tokenizer = tokenizer
        self.vocabulary = vocabulary
        self.max_length_question: int = 0
        self.max_length_context: int = 0
        self.is_fitted = False

    def fit_to_dataset(
        self, dataset: Dataset[Tuple[str, str, List[str], List[int]]]
    ) -> None:
        """Method to fit the vocabulary to the input text

        Args:
            dataset:
                the dataset to fit to

        """
        self.counter = Counter()
        self.ordered_dict = OrderedDict()
        for (context, question, answer, init_position) in dataset:  # type: ignore
            self.counter.update(self.tokenizer(context))  # type: ignore
            self.max_length_context = max(self.max_length_context, len(self.tokenizer(context)))  # type: ignore
            self.counter.update(self.tokenizer(question))  # type: ignore
            self.max_length_question = max(self.max_length_question, len(self.tokenizer(question)))  # type: ignore

        if self.vocabulary is None:
            self.ordered_dict = OrderedDict(
                sorted(self.counter.items(), key=lambda x: x[1], reverse=True)
            )
            self.vocabulary = vocab(self.ordered_dict)
            unk_token = "<unk>"  # type: ignore
            if unk_token not in self.vocabulary:
                self.vocabulary.insert_token(unk_token, 0)
            self.vocabulary.set_default_index(self.vocabulary[unk_token])
        self.pad_item = 0
        self.is_fitted = True

    def __call__(
        self, datum: Tuple[str, str, List[str], List[int]]
    ) -> Tuple[Union[Tensor, List[Tensor]], Tensor]:
        """This method implement the transformation once fitted."""
        # if not self.is_fitted:
        #    self.load_pretrained(".")
        if self.vocabulary:
            text_pipeline: Callable[[str], List[int]] = lambda x: [
                self.vocabulary[token] for token in self.tokenizer(x)  # type: ignore
            ]  # type: ignore
        else:
            raise MissingVocabularyError(
                "Please fit this preprocessor to initialise the vocabulary"
            )

        processed_context = torch.tensor(text_pipeline(datum[0]), dtype=torch.int64).to(
            DEVICE
        )
        out_context = torch.cat(
            [
                processed_context,
                self.pad_item
                * torch.ones(self.max_length_context - processed_context.shape[0]).to(
                    DEVICE
                ),
            ]
        )
        processed_question = torch.tensor(
            text_pipeline(datum[1]), dtype=torch.int64
        ).to(DEVICE)

        out_question = torch.cat(
            [
                processed_question,
                self.pad_item
                * torch.ones(self.max_length_question - processed_question.shape[0]).to(
                    DEVICE
                ),
            ]
        )

        pos_init_char = datum[3][0]
        pos_init = len(self.tokenizer(datum[0][:pos_init_char]))  # type: ignore
        pos_end = pos_init + len(self.tokenizer(datum[2][0]))  # type: ignore
        out_list = [out_context.to(torch.long), out_question.to(torch.long)]
        return out_list, torch.stack(
            (torch.tensor(pos_init), torch.tensor(pos_end))
        ).to(torch.long)



import os
import torch

# Define the root directory of the project which is parent of the parent of
# the current directory
ROOT_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), os.pardir, os.pardir)
)

# Define the default data directory
DEFAULT_DATA_DIR = os.path.join(ROOT_DIR, "examples", "data")

# Define the default dataset download directory
DEFAULT_DOWNLOAD_DIR = os.path.join(ROOT_DIR, "examples", "data", "DatasetCloud")

# Define the default dataset bucket on Google Cloud Storage where the datasets
# are stored
DATASET_BUCKET_NAME = "adversarial_attack"


# Define the default dataset download directory where the graph
# datasets from the PyG (PyTorch Geometric) library are stored
DEFAULT_GRAPH_DIR = os.path.join(ROOT_DIR, "examples", "data", "GraphDatasets")

# Define the default device: "cuda" is available, "cpu" otherwise
DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")



from typing import List, Optional, Union, Tuple, Any
import random
from copy import copy
import warnings

import torch
from torchvision.utils import make_grid
from sklearn.manifold import MDS
from gtda.diagrams import BettiCurve
from html2image import Html2Image
from captum.attr import visualization
import numpy as np
from captum.attr import LayerAttribution
from gtda.plotting import plot_diagram, plot_betti_surfaces, plot_betti_curves
import matplotlib.pyplot as plt

from gdeep.trainer import Trainer
from gdeep.analysis.interpretability import Interpreter
from gdeep.utility import DEVICE
from ..models import ModelExtractor
from . import (
    persistence_diagrams_of_activations,
    _simplified_persistence_of_activations,
    plotly2tensor,
    Compactification,
    png2tensor,
)

FONT_SIZE = 16
from gdeep.utility.custom_types import Tensor

from gdeep.utility.custom_types import Array


class Visualiser:
    """This class is the bridge to send to the tensorboard
    the required plots. It converts plots and images of
    different kinds to tensors

    Args:
        pipe :
            the Trainer instance to get info from

    Examples::

        from gdeep.visualization import Visualiser
        # you must have an initialized `trainer`
        vs = Visualiser(trainer)
        vs.plot_data_model()

    """

    persistence_diagrams: Optional[List[Array]] = None

    def __init__(self, pipe: Trainer) -> None:
        self.pipe = pipe

    def plot_interactive_model(self) -> None:
        """This function has no arguments: its purpose
        is to store the model to tensorboard for an
        interactive visualization.
        """

        data_iter = iter(self.pipe.dataloaders[0])
        x, _ = next(data_iter)
        new_x: List[Tensor] = []
        if isinstance(x, tuple) or isinstance(x, list):
            for i, xi in enumerate(x):
                new_x.append(xi.to(DEVICE))
            x = copy(new_x)
        else:
            x = x.to(DEVICE)
        # add interactive model to tensorboard
        self.pipe.writer.add_graph(self.pipe.model, x)  # type: ignore
        self.pipe.writer.flush()  # type: ignore

    def plot_3d_dataset(self, n_pts: int = 100) -> None:
        """This function has no arguments: its purpose
        is to store data to the tensorboard for
        visualization. Note that we are expecting the
        dataset item to be of type Tuple[Tensor, Tensor]

        Args:
            n_pts:
                number of points to display
        """
        data_iter = iter(self.pipe.dataloaders[0])
        features_list: List[Tensor] = []
        labels_list: List[str] = []

        for img, lab in data_iter:  # loop over batches
            if isinstance(img, tuple) or isinstance(img, list):
                features_list = features_list + [img[i][0] for i in range(len(img))]
            else:
                features_list = features_list + [img[i] for i in range(len(img))]
            try:
                labels_list = labels_list + [
                    str(lab[i].item()) for i in range(len(lab))
                ]
            except ValueError:
                labels_list = labels_list + [
                    "Label not available" for _ in range(len(img))
                ]
            if len(features_list) >= n_pts:
                break
        max_number = len(features_list)  # effective number of points

        features = torch.cat(features_list).to(DEVICE)

        if len(features.shape) >= 3:
            if len(features.shape) == 3:
                features = features.reshape(
                    max_number, -1, features.shape[1], features.shape[2]
                )
                grid = make_grid(features)
            else:
                grid = make_grid(features)

            self.pipe.writer.add_image("dataset", grid, 0)  # type: ignore

            self.pipe.writer.add_embedding(  # type: ignore
                features.view(max_number, -1),
                metadata=labels_list,
                label_img=features,
                tag="dataset",
                global_step=0,
            )
        else:
            self.pipe.writer.add_embedding(  # type: ignore
                features.view(max_number, -1),
                metadata=labels_list,
                tag="dataset",
                global_step=0,
            )

        self.pipe.writer.flush()  # type: ignore

    def plot_activations(
        self, batch: Optional[List[Union[List[Tensor], Tensor]]] = None
    ) -> None:
        """Plot PCA of the activations of all layers of the
        `self.model`.

        Args:
            batch:
                this should be an input batch for the model
        """
        me = ModelExtractor(self.pipe.model, self.pipe.loss_fn)
        if batch is not None:
            inputs, labels = batch
        else:
            inputs, labels = next(iter(self.pipe.dataloaders[0]))

        acts = me.get_activations(inputs)
        print("Sending the plots to tensorboard: ")
        for i, act in enumerate(acts):
            print("Step " + str(i + 1) + "/" + str(len(acts)), end="\r")
            length = act.shape[0]
            self.pipe.writer.add_embedding(  # type: ignore
                act.view(length, -1),
                metadata=labels,
                tag="activations_" + str(i),
                global_step=0,
            )
            self.pipe.writer.flush()  # type: ignore

    def plot_persistence_diagrams(
        self,
        batch: Optional[List[Tensor]] = None,
        homology_dimensions: Optional[List[int]] = None,
        **kwargs
    ) -> None:  # type: ignore
        """Plot a persistence diagrams of the activations

        Args:
            batch:
                a batch of data, in the shape of (datum, label)
            homology_dimensions:
                a list of the homology dimensions, like ``[0, 1]``
            kwargs:
                arguments for the ``persistence_diagrams_of_activations``
        """

        if homology_dimensions is None:
            homology_dimensions = [0, 1]
        me = ModelExtractor(self.pipe.model, self.pipe.loss_fn)
        if self.persistence_diagrams is None:
            if batch is not None:
                inputs, _ = batch
            else:
                inputs, _ = next(iter(self.pipe.dataloaders[0]))

            activation = me.get_activations(inputs)
            self.persistence_diagrams = persistence_diagrams_of_activations(
                activation, homology_dimensions=homology_dimensions, **kwargs
            )
        list_of_dgms = []
        for i, persistence_diagram in enumerate(self.persistence_diagrams):
            plot_persistence_diagram = plot_diagram(persistence_diagram)
            img_t = plotly2tensor(plot_persistence_diagram)
            list_of_dgms.append(img_t)
        features = torch.stack(list_of_dgms)
        # grid = make_grid(features)
        self.pipe.writer.add_images(  # type: ignore
            "Persistence diagrams of the activations", features, dataformats="NHWC"
        )
        self.pipe.writer.flush()  # type: ignore

    def plot_decision_boundary(
        self, compact: bool = False, n_epochs: int = 100, precision: float = 0.1
    ):
        """Plot the decision boundary as the intrinsic
        hypersurface defined by ``self.loss_fn == 0.5``
        in the case of 2 classes. The method also works
        for an arbitrary number of classes. Note that
        we are expecting a model whose forward function
        has only one tensor as input (and not multiple
        arguments).

        Args:
            compact:
                if plotting the compactified
                version of the boundary
            n_epochs:
                number of training cycles to find
                the decision boundary
            precision:
                parameter to filter the spurious data that are
                close to te decision boundary, but not close enough

        """

        me = ModelExtractor(self.pipe.model, self.pipe.loss_fn)
        x = next(iter(self.pipe.dataloaders[0]))[0][0]

        if compact:
            # initialization of the compactification
            cc = Compactification(
                neural_net=self.pipe.model,
                precision=0.1,
                n_samples=500,
                epsilon=0.051,
                n_features=x.shape[0],  # if x is not List[Tensor]
                n_epochs=100,
            )

            d_final, label_final = cc.create_final_distance_matrix()
            embedding = MDS(n_components=3, dissimilarity="precomputed")
            db = embedding.fit_transform(d_final)
            self.pipe.writer.add_embedding(  # type: ignore
                db, tag="compactified_decision_boundary", global_step=0
            )
            self.pipe.writer.flush()  # type: ignore
            return db, d_final, label_final
        else:
            db = me.get_decision_boundary(x, n_epochs, precision)
            if len(db) > 0:
                self.pipe.writer.add_embedding(db, tag="decision_boundary", global_step=0)  # type: ignore
                self.pipe.writer.flush()  # type: ignore
            else:
                warnings.warn("No points sampled over the decision boundary")
            return db.cpu(), None, None

    def plot_betti_surface_layers(
        self,
        homology_dimensions: Optional[List[int]] = None,
        batch: Optional[Tensor] = None,
        **kwargs
    ) -> List[Any]:  # type: ignore
        """
        Args:
            homology_dimensions :
                A list of
                homology dimensions, e.g. ``[0, 1, ...]``
            batch :
                the selected batch of data to
                compute the activations on. By
                defaults, this method takes the
                first batch of elements
            kwargs:
                optional arguments for the creation of
                persistence diagrams

        """
        if homology_dimensions is None:
            homology_dimensions = [0, 1]
        if self.persistence_diagrams is None:
            me = ModelExtractor(self.pipe.model, self.pipe.loss_fn)
            if batch is not None:
                inputs, _ = batch
            else:
                inputs = next(iter(self.pipe.dataloaders[0]))  # type: ignore
            self.persistence_diagrams = persistence_diagrams_of_activations(
                me.get_activations(inputs),
                homology_dimensions=homology_dimensions,
                **kwargs
            )
        bc = BettiCurve()
        dgms = bc.fit_transform(self.persistence_diagrams)

        plots = plot_betti_surfaces(
            dgms, samplings=bc.samplings_, homology_dimensions=homology_dimensions
        )

        return plots  # type: ignore

    def plot_betti_curves_layers(
        self,
        homology_dimensions: Optional[List[int]] = None,
        batch: Optional[Tensor] = None,
        **kwargs
    ) -> None:  # type: ignore
        """
        Args:
            homology_dimensions :
                A list of
                homology dimensions, e.g. ``[0, 1, ...]``
            batch :
                the selected batch of data to
                compute the activations on. By
                defaults, this method takes the
                first batch of elements
            kwargs:
                optional arguments for the creation of
                persistence diagrams

        """
        if homology_dimensions is None:
            homology_dimensions = [0, 1]
        if self.persistence_diagrams is None:
            me = ModelExtractor(self.pipe.model, self.pipe.loss_fn)
            if batch is not None:
                inputs, _ = batch
            else:
                inputs = next(iter(self.pipe.dataloaders[0]))  # type: ignore
            self.persistence_diagrams = persistence_diagrams_of_activations(
                me.get_activations(inputs),
                homology_dimensions=homology_dimensions,
                **kwargs
            )
        bc = BettiCurve()
        bc_curves = bc.fit_transform(self.persistence_diagrams)
        list_of_plts = []
        for bc_curve in bc_curves:
            plots = plot_betti_curves(
                bc_curve,
                samplings=bc.samplings_,
                homology_dimensions=homology_dimensions,
            )
            img_t = plotly2tensor(plots)
            list_of_plts.append(img_t)
        features = torch.stack(list_of_plts)
        self.pipe.writer.add_images(  # type: ignore
            "Betti curves for each layer", features, dataformats="NHWC"
        )
        self.pipe.writer.flush()  # type: ignore

    def plot_interpreter_text(self, interpreter: Interpreter):
        """This method allows to plot the results of an
        Interpreter for text data.

        Args:
            interpreter :
                this is a ``gdeep.analysis.interpretability``
                initialized ``Interpreter`` class

        Returns:
            matplotlib.figure
        """
        viz = interpreter.stored_visualizations
        fig = visualization.visualize_text(viz)
        name = "out.png"
        hti = Html2Image()
        hti.screenshot(html_str=fig.data, save_as=name)
        img_ten = png2tensor(name)
        self.pipe.writer.add_image(interpreter.method, img_ten, dataformats="HWC")  # type: ignore
        return fig

    def plot_interpreter_image(self, interpreter: Interpreter):
        """This method allows to plot the results of an
        Interpreter for image data.

        Args:
            interpreter:
                this is a ``gdeep.analysis.interpretability``
                initialized ``Interpreter`` class

        Returns:
            matplotlib.figure
        """

        try:
            attrib = (
                torch.permute(interpreter.attribution.squeeze().detach(), (1, 2, 0))
                .detach()
                .cpu()
                .numpy()
            )
        except ValueError:
            attrib = torch.permute(
                LayerAttribution.interpolate(
                    interpreter.attribution.detach().cpu(),
                    tuple(interpreter.x.squeeze().detach().cpu().shape[-2:]),
                ).squeeze(0),
                (1, 2, 0),
            )
            attrib = (
                torch.stack([attrib, attrib, attrib], dim=2)
                .squeeze(-1)
                .detach()
                .cpu()
                .numpy()
            )
        img = (
            torch.permute(interpreter.x.squeeze().detach(), (1, 2, 0))
            .detach()
            .cpu()
            .numpy()
        )

        fig, _ = visualization.visualize_image_attr_multiple(
            attrib,
            img,
            ["original_image", "heat_map", "blended_heat_map"],
            ["all", "all", "all"],
            show_colorbar=True,
        )
        self.pipe.writer.add_figure(interpreter.method, fig)  # type: ignore
        return fig

    def plot_feature_importance(self, interpreter: Interpreter):
        """This method allows to plot the results of an
        Interpreter for tabular data.

        Args:
            interpreter :
                this is a ``gdeep.analysis.interpretability``
                initialized ``Interpreter`` class

        Returns:
            matplotlib.figure
        """
        # prepare attributions for visualization
        x_test = interpreter.x
        x_axis_data = np.arange(x_test.shape[1])
        x_axis_data_labels = list(map(lambda idx: idx, x_axis_data))

        width = 0.14
        legends = interpreter.features_list

        fig = plt.figure(figsize=(20, 10))

        ax = plt.subplot()
        ax.set_title(
            "Comparing input feature importance across "
            + "multiple algorithms and learned weights"
        )
        ax.set_ylabel("Attributions")

        plt.rc("font", size=FONT_SIZE)  # fontsize of the text sizes
        plt.rc("axes", titlesize=FONT_SIZE)  # fontsize of the axes title
        plt.rc("axes", labelsize=FONT_SIZE)  # fontsize of the x and y labels
        plt.rc("legend", fontsize=FONT_SIZE - 4)  # fontsize of the legend
        for i in range(len(interpreter.attribution_list)):  # type: ignore
            attribution_sum = interpreter.attribution_list[i].detach().cpu().numpy().sum(0)  # type: ignore
            attribution_norm_sum = attribution_sum / np.linalg.norm(
                attribution_sum, ord=1
            )
            r = lambda: random.randint(0, 255)  # noqa
            color = "#%02X%02X%02X" % (r(), r(), r())
            ax.bar(
                x_axis_data,
                attribution_norm_sum,
                width,
                align="center",
                alpha=0.8,
                color=color,
            )

        ax.autoscale_view()
        plt.tight_layout()

        ax.set_xticks(x_axis_data + 0.5)
        ax.set_xticklabels(x_axis_data_labels)

        plt.legend(legends, loc=3)
        self.pipe.writer.add_figure("Feature Importance", fig)  # type: ignore
        return plt

    def plot_attribution(
        self, interpreter: Interpreter, **kwargs
    ) -> Tuple[plt.Figure, plt.Figure]:  # type: ignore
        """this method generically plots the attribution of
        the interpreter

        Args:
            interpreter :
                this is a ``gdeep.analysis.interpretability``
                initialized ``Interpreter`` class
            kwargs:
                keywords arguments for the plot (visualize_image_attr of
                captum)

        Returns:
            matplotlib.figure, matplotlib.figure
                they correspond respectively to the datum and attribution"""
        t1 = Visualiser._adjust_tensors_to_plot(interpreter.x)
        t2 = Visualiser._adjust_tensors_to_plot(interpreter.attribution)
        fig1, _ = visualization.visualize_image_attr(t1, **kwargs)
        fig2, _ = visualization.visualize_image_attr(t2, **kwargs)
        self.pipe.writer.add_figure("Datum x", fig1)  # type: ignore
        self.pipe.writer.add_figure("Generic attribution of x", fig2)  # type: ignore
        return fig1, fig2

    def plot_self_attention(
        self,
        attention_tensor: List[Tensor],
        tokens_x: Optional[List[str]] = None,
        tokens_y: Optional[List[str]] = None,
        **kwargs
    ) -> plt.Figure:  # type: ignore
        """This functions plots the self-attention layer of a transformer.

        Args:
            attention_tensor:
                list of the self-attention tensors (i.e. the tensors
                corresponding to the activations, given an input
            tokens_x:
                The string tokens to be displayed along the
                x axis of the plot
            tokens_y:
                The string tokens to be displayed along the
                y axis of the plot

        Returns:
             Figure:
                 matplotlib pyplot
        """
        fig = plt.figure(**kwargs)

        for idx, scores in enumerate(attention_tensor):
            scores_np = Visualiser._adjust_tensors_to_plot(scores)
            ax = fig.add_subplot(4, 3, idx + 1)
            # append the attention weights
            im = ax.imshow(scores_np, cmap="viridis")

            fontdict = {"fontsize": 10}
            if tokens_x:
                ax.set_xticks(range(len(tokens_x)))
                ax.set_xticklabels(tokens_x, fontdict=fontdict, rotation=90)
            if tokens_y:
                ax.set_yticks(range(len(tokens_y)))
                ax.set_yticklabels(tokens_y, fontdict=fontdict)

            ax.set_xlabel("Head {}".format(idx + 1))

            fig.colorbar(im, fraction=0.046, pad=0.04)
        # plt.show()
        self.pipe.writer.add_figure("Self attention map", fig)  # type: ignore
        self.pipe.writer.flush()  # type: ignore
        return fig

    def plot_attributions_persistence_diagrams(
        self, interpreter: Interpreter, **kwargs
    ) -> plt.Figure:  # type: ignore
        """this method allows the plot, on top of the persistence
        diagram, of the attribution values. For example, this would
        be the method to call when you run saliency maps on the
        persistence diagrams. Note that all homology dimensions
        are plot together

        Args:
            interpreter :
                this is a ``gdeep.analysis.interpretability``
                initialized ``Interpreter`` class
            kwargs:
                keywords arguments for the plot (``matplotlib.pyplot``)

        Returns:
            matplotlib.figure.Figure
        """

        pd_x: Array = interpreter.x.detach().cpu().numpy()
        pd_attr: Array = interpreter.attribution.detach().cpu().numpy()

        yy: Array = (np.abs(pd_attr.T[0]) + np.abs(pd_attr.T[1])).flatten()
        yy = (yy - yy.min(axis=0)) / (yy.max(axis=0) - yy.min(axis=0))
        fig = plt.figure(figsize=(10, 10))
        plt.scatter(x=pd_x.T[0], y=pd_x.T[1], c=yy, **kwargs)
        line = np.arange(0, max(pd_x.T[1]) * 1.1, 0.01)
        plt.plot(line, line, "--", color="black")
        plt.xlabel("Birth")
        plt.ylabel("Death")

        self.pipe.writer.add_figure("Persistence diagrams attributions", fig)  # type: ignore
        self.pipe.writer.flush()  # type: ignore

        return fig

    def plot_betti_numbers_layers(
        self,
        homology_dimensions: Optional[List[int]] = None,
        filtration_value: float = 1.0,
        batch: Optional[Tensor] = None,
        **kwargs
    ) -> plt.Figure:  # type: ignore
        """
        Args:
            homology_dimensions :
                A list of
                homology dimensions, e.g. ``[0, 1, ...]``
            filtration_value:
                the filtration value you want to threshold the
                filtration at
            batch :
                the selected batch of data to
                compute the activations on. By
                defaults, this method takes the
                first batch of elements
            kwargs:
                optional arguments for the creation of
                persistence diagrams

        """
        if homology_dimensions is None:
            homology_dimensions = [0, 1]

        me = ModelExtractor(self.pipe.model, self.pipe.loss_fn)
        if batch is not None:
            inputs, _ = batch
        else:
            inputs = next(iter(self.pipe.dataloaders[0]))  # type: ignore

        simplified_persistence_diagrams = _simplified_persistence_of_activations(
            me.get_activations(inputs), homology_dimensions, filtration_value, **kwargs
        )

        betti_numbers = []

        for dgm in simplified_persistence_diagrams:
            xx = dgm[:, 2]
            betti_numbers.append(
                [
                    len(xx[(dgm[:, 2] == dim) * (dgm[:, 1] == 1.0)])
                    for dim in homology_dimensions
                ]
            )
        betti_array = np.array(betti_numbers).T

        fig = plt.figure(figsize=(10, 6.8))
        for i, array in enumerate(betti_array):
            plt.plot(array, label=str(i) + "-th betti number")
        plt.xlabel("Layers")
        plt.ylabel("Betti numbers")
        plt.legend()

        self.pipe.writer.add_figure("Betti numbers over the layers", fig)  # type: ignore
        self.pipe.writer.flush()  # type: ignore

        return fig

    @staticmethod
    def _adjust_tensors_to_plot(tensor: Tensor) -> np.ndarray:
        """make sure that tensors that will be plotted as images have the
        correct dimensions"""
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, :]
        elif len(tensor.shape) > 4:
            tensor = tensor[0, :, :, :, 0]
        list_of_permutation = torch.argsort(torch.tensor(tensor.shape)).tolist()
        list_of_permutation.reverse()
        tensor = tensor.permute(*list_of_permutation)

        if tensor.shape[-1] == 2:
            temporary = torch.zeros((tensor.shape[0], tensor.shape[1], 3))
            temporary[:, :, :2] = tensor
            tensor = temporary
        else:
            tensor = tensor[:, :, :4]
        return tensor.permute(1, 0, 2).cpu().detach().numpy()



from gdeep.trainer import TrainerConfig
from torch.optim import SGD


def test_trainer_config() -> None:
    config = TrainerConfig(SGD)
    dictionary = config.to_dict()
    assert isinstance(dictionary, dict)



import torch
import torch.nn as nn
from torch.utils.data.sampler import SubsetRandomSampler
from torch.utils.tensorboard.writer import SummaryWriter

from gdeep.data.preprocessors import TokenizerTextClassification
from gdeep.data.datasets import DataLoaderBuilder, DatasetBuilder
from gdeep.trainer import Trainer
from gdeep.utility import DEVICE
from gdeep.visualization import Visualiser


bd = DatasetBuilder(name="AG_NEWS", convert_to_map_dataset=True)
ds_tr_str, ds_val_str, ds_ts_str = bd.build()


ptd = TokenizerTextClassification()

ptd.fit_to_dataset(ds_tr_str)
transformed_textds = ptd.attach_transform_to_dataset(ds_tr_str)  # type: ignore
transformed_textts = ptd.attach_transform_to_dataset(ds_val_str)  # type: ignore

# the only part of the training/test set we are interested in
train_indices = list(range(64 * 10))
test_indices = list(range(64 * 5))

dl_tr2, dl_ts2, _ = DataLoaderBuilder([transformed_textds, transformed_textts]).build(
    [
        {"batch_size": 16, "sampler": SubsetRandomSampler(train_indices)},
        {"batch_size": 16, "sampler": SubsetRandomSampler(test_indices)},
    ]
)
writer = SummaryWriter()


class TextClassificationModel(nn.Module):
    def __init__(self, vocab_size, embed_dim, num_class):
        super(TextClassificationModel, self).__init__()
        self.embedding = nn.Embedding(vocab_size, embed_dim, sparse=True)
        self.fc = nn.Linear(embed_dim, num_class)
        self.init_weights()

    def init_weights(self):
        init_range = 0.5
        self.embedding.weight.data.uniform_(-init_range, init_range)
        self.fc.weight.data.uniform_(-init_range, init_range)
        self.fc.bias.data.zero_()

    def forward(self, text):
        embedded = self.embedding(text)
        mean = torch.mean(embedded, dim=1)
        return self.fc(mean)


def test_visualiser():
    assert ptd.vocabulary is not None, "vocabulary is None"
    vocab_size = len(ptd.vocabulary)
    embedding_size = 64
    loss_fn = nn.CrossEntropyLoss()
    model = TextClassificationModel(vocab_size, embedding_size, 4).to(DEVICE)
    pipe = Trainer(model, [dl_tr2, dl_ts2], loss_fn, writer)

    vs = Visualiser(pipe)

    x = next(iter(dl_tr2))
    vs.plot_interactive_model()
    vs.plot_3d_dataset()
    vs.plot_activations(x)
    vs.plot_persistence_diagrams(x)
    vs.plot_betti_surface_layers([0, 1], x)
    vs.plot_betti_curves_layers([0, 1], x)
    vs.plot_betti_numbers_layers(batch=x, filtration_value=0.3)



from gdeep.visualization import Compactification
from gdeep.models import FFNet
from gdeep.utility import DEVICE


def test_compactification():
    model = FFNet((2, 4, 2)).to(DEVICE)
    c = Compactification(
        model,
        0.4,
        2,
        1500,
        0.05,
        50,
        [(-1.0, 1.0), (-1.0, 1.0)],
    )
    try:
        c.create_final_distance_matrix()
        c.plot_chart(1)
    except ValueError:
        pass  # this is because the charts are empty of points



from typing import Any, Callable, Generic, Iterable, TypeVar

from torch.utils.data import Dataset

from .abstract_preprocessing import AbstractPreprocessing
from .transforming_dataset import TransformingDataset, append_transform

T = TypeVar("T")


class PreprocessingPipeline(AbstractPreprocessing[T, Any], Generic[T]):
    """ Pipeline to fit non-fitted preprocessors to a dataset in a sequential manner.
    The fitted preprocessing transform can be attached to a dataset using
    the ´attach_transform_to_dataset´ method.
    The intended use case is to fit the preprocessors to the training dataset and
    then attach the fitted transform to the training, validation and test datasets.

    The transform is only applied to the data and not the labels.

    Examples::

        from gdeep.data.preeprocessors import PreprocessingPipeline, Normalization, \
            PreprocessImageClassification
        from gdeep.data.datasets import DatasetImageClassificationFromFiles

        image_dataset = DatasetImageClassificationFromFiles(
            os.path.join(file_path, "img_data"),
            os.path.join(file_path, "img_data", "labels.csv"))

        preprocessing_pipeline = PreprocessingPipeline((PreprocessImageClassification((32, 32)),
                                                        Normalization()))
        preprocessing_pipeline.fit_to_dataset(image_dataset)  # this will not change the image_dataset
        preprocessed_dataset = preprocessing_pipeline.attach_transform_to_dataset(image_dataset)

    """

    transform_composition: Callable[[T], Any]

    def __init__(
        self, preprocessors: Iterable[AbstractPreprocessing[Any, Any]]
    ) -> None:
        self.preprocessors = preprocessors
        self.is_fitted = False

    def attach_transform_to_dataset(
        self, dataset: Dataset[T]
    ) -> TransformingDataset[T, Any]:
        return TransformingDataset(dataset, self.transform_composition)  # type: ignore

    def fit_to_dataset(self, dataset: Dataset[Any]) -> None:
        self.transform_composition = lambda x: x  # type: ignore
        transformed_dataset = TransformingDataset(dataset, self.transform_composition)  # type: ignore
        for preprocessor in self.preprocessors:
            preprocessor.fit_to_dataset(transformed_dataset)
            transformed_dataset = append_transform(transformed_dataset, preprocessor)
        self.transform_composition = transformed_dataset.transform  # type: ignore

    def __call__(self, x: T) -> Any:
        if not self.is_fitted:
            raise ValueError(
                "The preprocessing pipeline has not been fitted to a dataset."
                " Please call the ´fit_to_dataset´ method first."
            )
        return self.transform_composition(x)  # type: ignore



import logging
import os
from os import listdir, makedirs
from os.path import isfile, join, isdir, exists, getsize
import requests  # type: ignore
import sys
from typing import Optional, Union, List
import time

import google
from google.cloud import storage  # type: ignore
from google.oauth2 import service_account  # type: ignore
import wget  # type: ignore

from gdeep.utility.constants import DEFAULT_DOWNLOAD_DIR, DATASET_BUCKET_NAME
from gdeep.utility.utils import get_checksum

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.WARNING)


def _check_public_access(use_public_access: bool):
    """Check if the public access is enabled."""

    def wrap(function):
        def wrapper_function(*args, **kwargs):
            if use_public_access:
                raise ValueError("DataCloud object has public access only!")
            return function(*args, **kwargs)

        return wrapper_function

    return wrap


class _DataCloud:
    """Download handle for Google Cloud Storage buckets.

    Args:
        bucket_name (str, optional):
            Name of the Google Cloud Storage bucket.
            Defaults to "adversarial_attack".
        download_directory (str, optional):
            Directory of the downloaded files.
            Defaults to join('examples', 'data', 'DataCloud').
        use_public_access: (bool, optional):
            Whether or not to use public api access.
            Defaults to True.
        path_credentials (str, optional):
            Path to the credentials file.
            Only used if public_access is False and credentials are not
            provided. Defaults to None.

        Raises:
            ValueError: If the bucket does not exist.

        Returns:
            None
    """

    def __init__(
        self,
        bucket_name: str = DATASET_BUCKET_NAME,
        download_directory: str = DEFAULT_DOWNLOAD_DIR,
        use_public_access: bool = True,
        path_to_credentials: Union[str, None] = None,
    ) -> None:
        self.bucket_name = bucket_name
        self.use_public_access = use_public_access
        if not self.use_public_access:
            # Get storage client
            if path_to_credentials is None:
                self.storage_client = storage.Client()
            else:
                credentials = service_account.Credentials.from_service_account_file(
                    path_to_credentials
                )
                self.storage_client = storage.Client(credentials=credentials)
            self.bucket = self.storage_client.bucket(self.bucket_name)
        else:
            self.public_url = "https://storage.googleapis.com/" + bucket_name + "/"

        # Set up download path
        self.download_directory = download_directory

        # Create a new directory because it does not exist
        if not exists(self.download_directory) and self.download_directory != "":
            makedirs(self.download_directory)

    def list_blobs(self) -> List[str]:
        """List all blobs in the bucket.

        Returns:
            List[str]:
                List of blobs in the bucket.
        """
        # Assert that the bucket does not use public access
        if self.use_public_access:
            raise ValueError(
                "DataCloud object can only list blobs" "when using private access!"
            )
        blobs = self.bucket.list_blobs()
        return [blob.name for blob in blobs]

    def blob_exists(self, blob_name: str) -> bool:
        """Check if a Blob exists in the bucket.

        Args:
            blob_name (str):
                Name of the Blob to check.


        Returns:
            bool:
                True if the Blob exists, False otherwise.
        """
        if self.use_public_access:
            url = self.public_url + blob_name
            response = requests.head(url)
            return response.status_code == 200
        else:
            blob = self.bucket.blob(blob_name)
            return blob.exists()

    def download_file(
        self, blob_name: str, download_directory: Union[str, None] = None
    ) -> None:
        """Download a blob from Google Cloud Storage bucket.

        Args:
            source_blob_name (str):
                Name of the blob to download. The name is relative to the
                root of the bucket.
            download_directory (str, optional):
                Directory to download the blob to.

        Raises:
            ValueError:
                If the blob does not exist.

        Returns:
            None
        """
        url = ""
        if download_directory is None:
            download_directory = self.download_directory
        if self.use_public_access:
            url = self.public_url + blob_name
        # Check if blob exists
        if not self.blob_exists(blob_name):
            raise google.api_core.exceptions.NotFound(  # type: ignore
                "Blob {} does not exist!".format(blob_name)
            )

        # If the file exists, compare checksums
        if isfile(join(download_directory, blob_name)):
            # Get remote md5 checksum from url in base64 format.
            if self.use_public_access:
                response = requests.get(url, stream=True)
                response.raw.decode_content = True
                # Get remote md5 checksum from url in base64 format if it exists.
                checksum_remote = response.headers.get("Content-MD5")
            else:
                blob = self.bucket.blob(blob_name)
                checksum_remote = blob.md5_hash

            checksum_local = get_checksum(
                join(download_directory, blob_name), encoding="base64"
            )
            if checksum_remote is not None:
                if checksum_remote != checksum_local:
                    # Ask user if they want to download the file
                    answer = input(
                        f"File {join(download_directory, blob_name)} already"
                        + "exists and checksums don't match! "
                        + "Do you want to overwrite it? [y/N]"
                    )
                    if answer.lower() not in ["y", "yes"]:
                        return
                else:
                    print(
                        f"File {join(download_directory, blob_name)} "
                        + "already exists and checksums match! "
                        + "Skipping download."
                    )
                    return
            else:
                print(
                    f"File {join(download_directory, blob_name)} already"
                    + "exists and remote checksum is "
                    + "None! Downloading anyway."
                )
        # Download file
        print("Downloading file {} to {}".format(blob_name, download_directory))
        if self.use_public_access:
            wget.download(url, join(download_directory, blob_name))
        else:
            self.bucket.blob(blob_name).download_to_filename(
                join(download_directory, blob_name), checksum="md5"
            )

    def download_folder(self, blob_name: str) -> None:
        """Download a folder from Google Cloud Storage bucket.

        Warning: This function does not download empty subdirectories.

        Args:
            blob_name (str):
                Name of the blob folder to download. The name is relative to
                the root of the bucket.

        Raises:
            RuntimeError:
                If the folder does not exist.

        Returns:
            None
        """
        assert not self.use_public_access, (
            "Downloading folders is not" "is not supported with public" "access!"
        )
        # Get list of files in the blob
        blobs = self.bucket.list_blobs(prefix=blob_name)
        for blob in blobs:
            # Do not download subdirectories
            if blob.name.endswith("/"):
                continue
            file_split = blob.name.split("/")
            directory = "/".join(file_split[0:-1])
            if not exists(directory):
                makedirs(join(self.download_directory, directory), exist_ok=True)
            logging.getLogger().info("Downloading blob %s", blob.name)

            local_path = (
                blob.name.replace("/", "\\") if sys.platform == "win32" else blob.name
            )

            blob.download_to_filename(
                join(self.download_directory, local_path), checksum="md5"
            )

    def upload_file(
        self,
        source_file_name: str,
        target_blob_name: Union[str, None] = None,
        make_public: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Upload a local file to Google Cloud Storage bucket.

        Args:
            source_file_name (str):
                Filename of the local file to upload.
            target_blob_name (Union[str, None], optional):
                Name of the target blob relative to the root of the bucket.
                If None, the filename will be used.
                Defaults to None.
            make_public (bool, optional):
                Whether or not to make the uploaded
                file public. Defaults to False.
                overwrite (bool, optional):
            Whether or not to overwrite the target
                Blob. Defaults to False.

        Raises:
            RuntimeError: If the target Blob already exists.

        Returns:
            None
        """
        if target_blob_name is None:
            target_blob_name = os.path.basename(source_file_name)
        blob = self.bucket.blob(target_blob_name)
        if blob.exists() and not overwrite:
            raise RuntimeError(f"Blob {target_blob_name} already exists.")
        logging.getLogger().info("upload file %s", source_file_name)
        # Check if source_file_name is bigger than 5GB
        if isfile(source_file_name) and getsize(source_file_name) > 5000000000:
            raise ValueError("File is bigger than 5GB")

        # Compute MD5 checksum of the file and add it to the metadata of
        # the blob
        blob.md5_hash = get_checksum(source_file_name, encoding="base64")

        blob.upload_from_filename(source_file_name, checksum="md5")
        if make_public:
            blob.make_public()

    def upload_folder(
        self,
        source_folder: str,
        target_folder: Optional[str] = None,
        make_public: bool = False,
    ) -> None:
        """Upload a local folder with all it's subolders to Google
        Cloud Storage bucket.

        Args:
            source_folder (str):
                Folder to upload.
            target_folder (Union[str, None], optional):
                Name of the target folder relative to the root of the bucket.
                If None, the root of the bucket will be used.
                Defaults to None.
            make_public (bool, optional):
                Whether or not to make the uploaded
                file public. Defaults to False.

        Raises:
            ValueError:
                If the source folder is not a directory.

        Returns:
            None
        """
        if not isdir(source_folder):
            raise ValueError("Source folder is not a directory.")

        if target_folder is None:
            target_folder = ""

        # List of all files in the source folder
        files = [
            join(source_folder, f)
            for f in listdir(source_folder)
            if isfile(join(source_folder, f))
        ]
        print(files)
        # Upload all files in the source folder
        for file in files:
            file_name = os.path.basename(file)
            if target_folder == "":
                self.upload_file(
                    join(source_folder, file_name),
                    target_blob_name=file_name,
                    make_public=make_public,
                )
            else:
                self.upload_file(
                    join(source_folder, file_name),
                    target_blob_name=target_folder + "/" + file_name,
                    make_public=make_public,
                )

        # List of all subfolders in the source folder
        subfolders = [
            join(source_folder, f)
            for f in listdir(source_folder)
            if isdir(join(source_folder, f))
        ]
        # Upload all subfolders in the source folder recursively
        for subfolder in subfolders:
            relative_subfolder = os.path.relpath(subfolder, source_folder)
            if target_folder == "":
                self.upload_folder(
                    join(source_folder, relative_subfolder),
                    target_folder=relative_subfolder,
                    make_public=make_public,
                )
            else:
                self.upload_folder(
                    join(source_folder, relative_subfolder),
                    target_folder=target_folder + "/" + relative_subfolder,
                    make_public=make_public,
                )

    def delete_blob(self, blob_name: str) -> None:
        """Deletes a single Blob from Google Cloud Storage

        Args:
            blob_name (str):
                The name of the Blob to delete

        Raises:
            RuntimeError: If the Blob does not exist.

        Returns:
            None
        """
        blob = self.bucket.blob(blob_name)
        blob.delete()

    def delete_blobs(self, blobs_name: str) -> None:
        """Deletes a Blob and all its children from Google Cloud Storage.

        Args:
            blobs_name (str):
                Name of the parent Blob to delete.

        Raises:
            ValueError:
                If the Blob does not exist.

        Returns:
            None
        """
        blobs = self.bucket.list_blobs(prefix=blobs_name)
        for blob in blobs:
            blob.delete()



import torch


class MissingClosureError(Exception):
    pass


class SAM(torch.optim.Optimizer):
    base_optimizer_cls = None

    def __init__(self, params, base_optimizer=None, rho=0.05, adaptive=False, **kwargs):
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"

        defaults = dict(rho=rho, adaptive=adaptive, **kwargs)
        super(SAM, self).__init__(params, defaults)
        if base_optimizer is not None:
            self.base_optimizer_cls = base_optimizer
        assert self.base_optimizer_cls is not None, "Must provide base_optimizer_cls"
        self.base_optimizer = self.base_optimizer_cls(self.param_groups, **kwargs)
        self.param_groups = self.base_optimizer.param_groups

    @torch.no_grad()
    def first_step(self, zero_grad=False):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)

            for p in group["params"]:
                if p.grad is None:
                    continue
                e_w = (
                    (torch.pow(p, 2) if group["adaptive"] else 1.0)
                    * p.grad
                    * scale.to(p)
                )
                p.add_(e_w)  # climb to the local maximum "w + e(w)"
                self.state[p]["e_w"] = e_w

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad=False):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                p.sub_(self.state[p]["e_w"])  # get back to "w" from "w + e(w)"

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def step(self, closure=None):
        if closure is None:
            raise MissingClosureError(
                "Sharpness Aware Minimization requires closure, but it was not provided"
            )
        closure = torch.enable_grad()(
            closure
        )  # the closure should do a full forward-backward pass

        self.first_step(zero_grad=True)
        closure()
        self.second_step()

    def _grad_norm(self):
        shared_device = self.param_groups[0]["params"][
            0
        ].device  # put everything on the same device, in case of model parallelism
        norm = torch.norm(
            torch.stack(
                [
                    ((torch.abs(p) if group["adaptive"] else 1.0) * p.grad)
                    .norm(p=2)
                    .to(shared_device)
                    for group in self.param_groups
                    for p in group["params"]
                    if p.grad is not None
                ]
            ),
            p=2,  # type: ignore
        )
        return norm


class SAMOptimizer(SAM):
    """This is the class to be used in the pipeline.

    First you need to initialise this class
    with the torch optimiser
    that you would like to apply SAM to.

    Then, at the following call, the instance
    behaves extcly like the SAM optimiser

    Args:
        optimizer (torch.optim):
            the optimiser class (not the instance)
    """

    def __new__(cls, optimizer):
        return type(
            "SAMOptimizer",
            (SAM, torch.optim.Optimizer),
            {"base_optimizer_cls": optimizer},
        )



#! /usr/bin/env python
"""Toolbox for Machine Learning using Topological Data Analysis."""

import os
import codecs

from setuptools import setup, find_packages


version_file = os.path.join("gdeep", "_version.py")
__version__ = "0.0.0"  # initialise the variable
with open(version_file) as f:
    exec(f.read())

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

DISTNAME = "giotto-deep"
DESCRIPTION = "Toolbox for Deep Learning and Topological Data Analysis."
with codecs.open("README.md", encoding="utf-8-sig") as f:
    LONG_DESCRIPTION = f.read()
LONG_DESCRIPTION_TYPE = "text/markdown"
MAINTAINER = "Matteo Caorsi"
MAINTAINER_EMAIL = "maintainers@giotto.ai"
URL = "https://github.com/giotto-ai/giotto-deep"
LICENSE = "GNU AGPLv3"
VERSION = __version__  # type: ignore
DOWNLOAD_URL = "https://github.com/giotto-ai/giotto-deep/tarball/v" + VERSION
CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]
KEYWORDS = (
    "deep learning, topological data analysis, persistent "
    "homology, persistence diagrams"
)
INSTALL_REQUIRES = requirements
EXTRAS_REQUIRE = {
    "tests": ["pytest", "flake8", "mypy"],
    "doc": [
        "openml",
        "sphinx",
        "nbconvert",
        "sphinx-issues",
        "sphinx_rtd_theme",
        "numpydoc",
    ],
    "examples": ["jupyter", "pandas", "openml", "plotly"],
}

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_TYPE,
    zip_safe=False,
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    keywords=KEYWORDS,
    dependency_links=["https://download.pytorch.org/whl/torch_stable.html"],
    find_links=["https://download.pytorch.org/whl/torch_stable.html"],
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
)



import warnings
from typing import Any, Tuple, TypeVar

import torch
from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)
from gdeep.utility import DEVICE
from matplotlib.pyplot import axis
from torch.utils.data import Dataset

from ..abstract_preprocessing import AbstractPreprocessing
from ..transforming_dataset import TransformingDataset
from .min_max_scalar import _compute_min_max_of_dataset

from gdeep.utility.custom_types import Tensor

PD = OneHotEncodedPersistenceDiagram

T = TypeVar("T")


class MinMaxScalarPersistenceDiagram(
    AbstractPreprocessing[Tuple[Tensor, T], Tuple[Tensor, T]]
):
    """This class runs the standard min-max normalisation on the birth and death times
    of the persistence diagrams. For example. The transformation is:
    X_scaled = X_std * (max - min) + min
    """

    is_fitted: bool
    min_max = Tuple[Tensor, Tensor]

    def __init__(self):
        self.is_fitted = False

    def fit_to_dataset(self, dataset: Dataset[Tuple[Tensor, T]]) -> None:
        self.min_max = _compute_min_max_of_dataset(
            TransformingDataset(
                dataset, lambda x: (torch.cat([x[0].min(), x[0].max()]), x[1])
            )
        )
        self.is_fitted = True

    def __call__(self, item: Tuple[Tensor, T]) -> Tuple[Tensor, T]:
        if not self.is_fitted:
            raise RuntimeError(
                "The normalisation is not fitted to any dataset."
                " Please call fit_to_dataset() first."
            )
        if not torch.all(self.min_max[1] > 0):  # type: ignore
            warnings.warn("The max values are zeros! Adding 1e-7")
            self.min_max = (self.min_max[0], self.min_max[1] + 1e-7)  # type: ignore
        out = (item[0] - self.min_max[0]) / self.min_max[1]  # type: ignore
        return (out, item[1])



from .one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
    get_one_hot_encoded_persistence_diagram_from_gtda,
    collate_fn_persistence_diagrams,
    get_one_hot_encoded_persistence_diagram_from_gudhi_extended,
)

__all__ = [
    "OneHotEncodedPersistenceDiagram",
    "collate_fn_persistence_diagrams",
    "get_one_hot_encoded_persistence_diagram_from_gtda",
    "get_one_hot_encoded_persistence_diagram_from_gudhi_extended",
]



from dataclasses import dataclass
from typing import Type, Optional, Dict, Any, Tuple

from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler  # noqa
from optuna.trial._base import BaseTrial  # noqa


@dataclass
class TrainerConfig:
    """Configuration class that contains the
    parameters of the Trainer and of the Benchmark class

    Args:
        optimizer:
            optimizer class, not the instance
        n_epochs:
            number of training epochs
        optimizers_params:
            dictionary of the optimizers
            parameters, e.g. `{"lr": 0.001}`
        dataloaders_params:
            dictionary of the dataloaders
            parameters
        models_hyperparams:
            dictionary of the model
            parameters
        lr_scheduler:
            a learning rate scheduler class
        schedulers_params:
            learning rate scheduler parameters
        optuna_params:
            a tuple with the optuna trial and the
            search metric (a string).
        profiling:
            whether or not you want to activate the
            profiler
        parallel_tpu:
            boolean value to run the computations
            on multiple TPUs
        keep_training:
            This flag allows to restart a training from
            the existing optimizer as well as the
            existing model
        store_grad_layer_hist:
            This flag allows to store the gradients
            and the layer values in tensorboard for
            each epoch
        n_accumulated_grads:
            this is the number of accumated grads. It
            is taken into account only for positive integers
        writer_tag:
            tag to prepend to the ouput
            on tensorboard
    """

    optimizer: Type[Optimizer]
    n_epochs: int = 10
    cross_validation: bool = False
    optimizers_param: Optional[Dict[str, Any]] = None
    dataloaders_param: Optional[Dict[str, Any]] = None
    lr_scheduler: Optional[Type[_LRScheduler]] = None
    scheduler_params: Optional[Dict[str, Any]] = None
    optuna_params: Optional[Tuple[BaseTrial, str]] = None
    profiling: bool = False
    parallel_tpu: bool = False
    keep_training: bool = False
    store_grad_layer_hist: bool = False
    n_accumulated_grads: int = 0
    writer_tag: str = ""

    def to_dict(self) -> Dict[str, Any]:
        """method to transform the config file into a dictionary"""
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}



import torch

from gdeep.utility.custom_types import Tensor


def accuracy(prediction: Tensor, y: Tensor) -> float:
    """This function computes the accuracy
    for the given prediction and expected
    output

    Args:
        prediction:
            the output of your model. If ``X`` is
            a tensor, then ``prediction = model(X)``
        y:
            the corresponding expected results

    Returns:
        float:
            the value of the accuracy
    """
    correct: float = 0.0
    try:
        correct += (prediction.argmax(1) == y).to(torch.float).sum().item()
    except RuntimeError:
        correct += (prediction.argmax(2) == y).to(torch.float).sum().item()

    return correct / y.shape[0] * 100



"""Testing for simpleNN."""
# License: GNU AGPLv3
import pytest
import torch
from .. import FFNet, PeriodicNeuralNetwork


def test_periodic():
    X = torch.rand(10, 2)
    sn = FFNet((2, 3, 2))
    bdry_list = [(-1, 1), (0.5, 1)]
    period = PeriodicNeuralNetwork(sn, bdry_list)
    out = period.forward(X)
    assert out.shape[0] == 10 and out.shape[1] == 2



from typing import Dict, Optional, Tuple, List
import os
from os.path import join

import torch
from torch.utils.data import DataLoader, Dataset

from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)


class PersistenceDiagramFromFiles(Dataset[Tuple[OneHotEncodedPersistenceDiagram, int]]):
    file_path: str
    lenght: int
    labels: Dict[int, int]

    def __init__(
        self,
        file_path: str,
    ):
        """
        Args:
            file_path:
                The path to the persistence diagrams.
                num_homology_types: The number of homology types.
                transform: A function that transforms the data.
        """
        self.file_path = file_path
        self.len = len(os.listdir(join(self.file_path, "diagrams")))
        self._load_labels()

    def __len__(self) -> int:
        """
        Return the length of the dataset.

        Returns:
            The length of the dataset.
        """
        return self.len

    def _load_labels(self) -> None:
        """
        Load the labels of the dataset.

        Returns:
            The labels of the dataset.
        """
        self.labels: Dict[int, int] = {}
        # load labels.csv with columns=["graph_idx", "label"] skipping the header
        with open(join(self.file_path, "labels.csv")) as f:
            for line in f:
                if line.startswith("graph_idx"):
                    continue
                graph_idx, label = line.strip().split(",")
                self.labels[int(graph_idx)] = int(label)
        self.labels

    def __getitem__(self, index: int) -> Tuple[OneHotEncodedPersistenceDiagram, int]:
        """
        Return the item at the specified index.

        Args:
            index:
                The index of the item.

        Returns:
            The item at the specified index.
        """
        diagram = OneHotEncodedPersistenceDiagram.load(
            join(self.file_path, "diagrams", f"{index}.npy")
        )
        label = self.labels[index]
        return diagram, label



from typing import Any, Tuple, Dict, Callable, List, Optional
import inspect

# from captum.attr import *
from captum.attr import TokenReferenceBase, visualization
import torch
from torch import nn
from .attribution_factory import get_attr

from gdeep.utility import DEVICE

from gdeep.utility.custom_types import Tensor


class Interpreter:
    """Class to visualise the activation maps,
    the attribution maps and saliency maps using
    different techniques.

    Args:
        model:
            the standard pytorch model
        method:
            the interpretability method. Find
            more info at https://captum.ai/tutorials/

    """

    x: Tensor
    attribution: Tensor
    sentence: str
    y: Tensor
    layer: Optional[torch.nn.Module]
    features_list: Optional[List]
    attribution_list: Optional[List]

    def __init__(self, model: nn.Module, method: str = "IntegratedGradients"):
        # self.model = model
        self.model = model.to(DEVICE)
        self.method = method
        self.stored_visualizations: List = []

    def interpret(
        self,
        x: Tensor,
        y: Optional[Any] = None,
        layer: Optional[torch.nn.Module] = None,
        **kwargs: Any
    ) -> Tuple[Tensor, Tensor]:
        """This method creates a datum interpreter. This class
        is based on captum.

        Args:
            x:
                the tensor corresponding to the input datum.
                In case the datum is an image for example,
                it is expected to be of size ``(b, c, h, w)``
            y:
                the label we want to check the interpretability
                of.
            layer (nn.Module, optional):
                some methods will require to specify a layer
                of self.model

        Returns
            torch.Tensor, torch.Tensor:
                the input datum and the attribution
                image respectively.
        """
        self.x = x.to(DEVICE)
        self.layer = layer
        if self.layer:
            attr_class = get_attr(self.method, self.model, self.layer)
        else:
            attr_class = get_attr(self.method, self.model)
        self.model.eval()
        if y is not None:
            self.attribution = attr_class.attribute(self.x, target=y, **kwargs)
        else:
            self.attribution = attr_class.attribute(self.x, **kwargs)
        return self.x, self.attribution

    def feature_importance(
        self, x: Tensor, y: Tensor, **kwargs: Any
    ) -> Tuple[Tensor, List[Any]]:
        """This method creates a tabular data interpreter. This class
        is based on captum.

        Args:
            x:
                the datum
            y:
                the target label
            kwargs:
                kwargs for the attributions

        Returns:
            (Tensor, Tensor):
                returns x and its attribution

        """
        self.x = x.to(DEVICE)  # needed for attribution functions
        self.y = y.to(DEVICE)
        self.model.eval()
        attr_list = []
        attribute_dict = {
            "inputs": self.x,
            "target": self.y,
            "sliding_window_shapes": (1,),
            "n_steps": 50,
            **kwargs,
        }
        self.features_list = [
            "IntegratedGradients",
            "IntegratedGradients",
            "DeepLift",
            "FeatureAblation",
            "Occlusion",
        ]
        for method in self.features_list:
            attribution_class = get_attr(method, self.model)
            args_names = inspect.signature(
                attribution_class.attribute
            ).parameters.keys()
            kwargs_of_method = {
                k: v for k, v in attribute_dict.items() if k in args_names
            }
            attr_list.append(attribution_class.attribute(**kwargs_of_method))
        self.attribution_list = attr_list
        return self.x, self.attribution_list

    def interpret_text(
        self,
        sentence: str,
        label: Any,
        vocab: Dict[str, int],
        tokenizer: Callable[[str], List[str]],
        layer: Optional[torch.nn.Module] = None,
        min_len: int = 7,
        **kwargs
    ) -> Tuple[str, Tensor]:
        """This method creates a text interpreter. This class
        is based on captum.

        Args:
            sentence :
                the input sentence
            label:
                the label we want to check the interpretability
                of.
            vocab :
                a ``gdeep.data.preprocessors`` vocabulary. Can
                be extracted via the ``vocabulary`` attribute.
            tokenizer :
                a ``gdeep.data.preprocessors`` tokenizer. Can
                be extracted via the ``tokenizer`` attribute.
            layer :
                torch module corresponding to the layer belonging to
                ``self.model``.
            min_len:
                minimum length of the text. Shorter texts are padded
            kwargs:
                additional arguments for the attribution

        """
        self.model.eval()
        self.sentence = sentence
        if layer:
            attr_class = get_attr(self.method, self.model, layer)
        else:
            attr_class = get_attr(self.method, self.model)
        text = tokenizer(sentence)
        if len(text) < min_len:
            text += [" "] * (min_len - len(text))
        indexed = [vocab[t] for t in text]

        self.model.zero_grad()

        input_indices = torch.tensor(indexed).to(DEVICE)
        input_indices = input_indices.unsqueeze(0)

        # input_indices dim: [sequence_length]
        seq_length = min_len

        # predict
        pred_temp = torch.softmax(self.model(input_indices), 1)
        pred = torch.max(pred_temp)
        pred_ind: float = torch.argmax(pred_temp).item()
        # generate reference indices for each sample
        pad_index = 0
        token_reference = TokenReferenceBase(reference_token_idx=pad_index)
        reference_indices = token_reference.generate_reference(
            seq_length, DEVICE
        ).unsqueeze(0)
        # compute attributions and approximation
        # delta using layer integrated gradients
        self.attribution, delta = attr_class.attribute(
            input_indices, reference_indices, target=label, **kwargs
        )

        print(
            "pred: ", pred_ind, "(", "%.2f" % pred, ")", ", delta: ", abs(delta.item())
        )

        self.add_attributions_to_visualizer(
            self.attribution,
            text,
            pred,
            pred_ind,
            label,
            delta.item(),
            self.stored_visualizations,
        )
        return sentence, self.attribution

    @staticmethod
    def add_attributions_to_visualizer(
        attributions,
        text: List[str],
        pred: Tensor,
        pred_ind: float,
        label: Any,
        delta: Any,
        vis_data_records: List,
    ) -> None:
        attributions = attributions.sum(dim=2).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.cpu().detach().numpy()

        # storing couple samples in an array for visualization purposes
        vis_data_records.append(
            visualization.VisualizationDataRecord(
                attributions,
                pred,
                str(pred_ind),
                str(label),
                "1",
                attributions.sum(),
                text,
                delta,
            )
        )



import torch
from ...optimization import PersistenceGradient


def test_persistence_gradient_2d():
    """check if the class is consistent."""
    X = torch.tensor([[1, 0.0], [0, 1.0], [2, 2], [2, 1]])
    hom_dim = [0, 1]
    pg = PersistenceGradient(
        homology_dimensions=hom_dim, zeta=0.1, max_edge_length=3, collapse_edges=False
    )

    assert X.dtype == torch.float32
    assert pg.phi(X).shape[0] == 14

    assert pg.persistence_function(X).item() >= -2.328427314758301 - 0.001
    pg.sgd(X, n_epochs=4, lr=0.4)


def test_persistence_gradient_3d():
    """check if the class is consistent."""
    X = torch.tensor([[1, 0.0, 1], [0, 1.0, 0], [2, 2, 1], [2, 1, 2]])
    hom_dim = [0, 1]
    pg = PersistenceGradient(
        homology_dimensions=hom_dim, zeta=0.1, max_edge_length=3, collapse_edges=False
    )
    assert X.dtype == torch.float32
    assert pg.phi(X).shape[0] == 14

    assert pg.persistence_function(X).item() >= -2.7783148288726807 - 0.001
    pg.sgd(X, n_epochs=4, lr=0.4)


def test_persistence_gradient_5d():
    """check if the class is consistent."""
    X = torch.tensor(
        [[1, 0.0, 1, 0.5, 1], [0, 1.0, 0, 0.5, 1], [2, 2, 1, 0.5, 1], [2, 1, 2, 0.5, 1]]
    )
    hom_dim = [0, 1]
    pg = PersistenceGradient(
        homology_dimensions=hom_dim, zeta=0.1, max_edge_length=3, collapse_edges=False
    )
    assert X.dtype == torch.float32
    assert pg.phi(X).shape[0] == 14
    assert pg.persistence_function(X).item() >= -2.7783148288726807 - 0.001
    pg.sgd(X, n_epochs=4, lr=0.4)


def test_persistence_gradient_4d():
    """check if the matrix input works properly"""
    X = torch.tensor(
        [[1, 0.0, 1, 0.5], [0, 1.0, 0, 0.5], [2, 2, 1, 0.5], [2, 1, 2, 0.5]]
    )
    hom_dim = [0, 1]
    pg = PersistenceGradient(
        homology_dimensions=hom_dim,
        zeta=0.1,
        max_edge_length=3,
        metric="precomputed",
        collapse_edges=False,
    )
    assert X.dtype == torch.float32
    assert pg.phi(X).shape[0] == 14
    assert pg.persistence_function(X).item() >= 0.3467579483985901 + 0.001
    pg.sgd(X)


def test_persistence_gradient_matrix():
    # simulate the weighted graph
    dist = torch.tensor([[0.0, 2, 3], [2, 0.0, 2.2], [3, 2.2, 0.0]])

    pg = PersistenceGradient(
        homology_dimensions=[0, 1], zeta=0.0, collapse_edges=False, metric="precomputed"
    )
    assert all(pg.phi(dist) == torch.tensor([0.0, 0.0, 0.0, 2.0, 2.2, 3.0, 3.0]))
    assert pg.persistence_function(dist).item() >= -6.3 - 0.0001
    pg.sgd(dist, n_epochs=1, lr=1)


def test_persistence_gradient_matrix_2():
    # simulate the weighted graph
    dist = torch.tensor(
        [
            [0.0, 2.0, 10.0, 10.0],
            [2.0, 0.0, 2.0, 1],
            [10.0, 2.0, 0.0, 1],
            [10.0, 1, 1, 0.0],
        ]
    )
    pg = PersistenceGradient(
        homology_dimensions=[0, 1], zeta=0.0, collapse_edges=False, metric="precomputed"
    )
    assert all(
        pg.phi(dist)
        == torch.tensor(
            [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 2.0, 2.0, 2.0, 10.0, 10.0, 10.0, 10.0, 10.0]
        )
    )
    assert pg.persistence_function(dist).item() > -23.0
    pg.sgd(dist, n_epochs=1, lr=0.002)


def test_persistence_gradient_pts():
    # test explicit gradients
    pts = torch.tensor([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0]])
    pg = PersistenceGradient(
        homology_dimensions=[0, 1], zeta=0.0, collapse_edges=False, metric="euclidean"
    )

    assert all(pg.phi(pts)[:5] == torch.tensor([0.0, 0.0, 0.0, 1.0, 1.0]))

    pg.sgd(pts, n_epochs=1, lr=0.002)

    assert (
        (pts.grad == torch.tensor([[1.0, 1.0], [0.0, -1.0], [-1.0, 0.0]])).all().item()
    )



import numpy as np

from ..parallel_orbit import OrbitsGenerator


parameters = (1.0, 2.0, 3.0)
num_orbits_per_class = 10
num_pts_per_orbit = 15

og = OrbitsGenerator(
    parameters=parameters,
    num_orbits_per_class=num_orbits_per_class,
    num_pts_per_orbit=num_pts_per_orbit,
)

orbits = og.get_orbits()


def test_size_of_orbits():
    """Tests if the size of the generator is correct"""
    assert orbits is not None, "Orbits are None"
    assert orbits.shape == (
        len(parameters) * num_orbits_per_class,
        num_pts_per_orbit,
        2,
    )


def test_size_of_persistence_diagrams():
    """Test if the size of the persistence diagram is correct"""

    persistence_diagrams = og.get_persistence_diagrams()

    assert persistence_diagrams is not None, "Persistence diagrams are None"
    assert [persistence_diagrams.shape[i] for i in (0, 2)] == [
        len(parameters) * num_orbits_per_class,
        2 + og._num_homology_dimensions,
    ]


def test_recursive_definition():
    """Tests if the recursively defined orbits are correct"""
    assert orbits is not None, "Orbits are None"
    for p_idx, p in enumerate(parameters):
        for o in range(num_orbits_per_class):
            for i in range(1, num_pts_per_orbit - 1):
                x_current = orbits[p_idx * num_orbits_per_class + o, i, 0]
                x_next = orbits[p_idx * num_orbits_per_class + o, i + 1, 0]
                y_current = orbits[p_idx * num_orbits_per_class + o, i, 1]
                y_next = orbits[p_idx * num_orbits_per_class + o, i + 1, 1]

                assert np.isclose(
                    x_next, (x_current + p * y_current * (1.0 - y_current)) % 1.0
                )
                assert np.isclose(
                    y_next, (y_current + p * x_next * (1.0 - x_next)) % 1.0
                )



import torch

from gdeep.trainer import accuracy


def test_accuracy():
    predictions = torch.rand((10, 3))
    y = torch.randint(0, 2, (10,))

    acc = accuracy(predictions, y)

    assert acc >= 0



import torch.nn as nn
from torch.nn import Dropout, Linear, Module, Sequential

from gdeep.topology_layers.pooling_layers import (
    AttentionPoolingLayer,
    MaxPoolingLayer,
    MeanPoolingLayer,
    SumPoolingLayer,
)
from gdeep.utility.enum_types import ActivationFunction, PoolerType
from .attention_factory import AttentionFactory
from .persformer_config import PersformerConfig


def get_pooling_layer(config: PersformerConfig) -> Module:
    """
    Get the pooling layer.

    Args:
        config: The configuration of the model.

    Returns:
        The pooling layer.
    """
    if config.pooler_type is PoolerType.ATTENTION:
        return AttentionPoolingLayer(config)  # type: ignore
    elif config.pooler_type is PoolerType.MAX:
        return MaxPoolingLayer(config)  # type: ignore
    elif config.pooler_type is PoolerType.MEAN:
        return MeanPoolingLayer(config)  # type: ignore
    elif config.pooler_type is PoolerType.SUM:
        return SumPoolingLayer(config)  # type: ignore
    else:
        raise ValueError(f"Pooler type {config.pooler_type} is not supported.")


def get_feed_forward_layer(config: PersformerConfig) -> Module:
    """
    Get the feed forward layer.
    """
    feed_forward_layer = Sequential()
    feed_forward_layer.add_module(
        "intermediate", Linear(config.hidden_size, config.intermediate_size)
    )
    feed_forward_layer.add_module(
        "activation", get_activation_function(config.hidden_act)
    )
    feed_forward_layer.add_module("dropout", Dropout(config.hidden_dropout_prob))

    feed_forward_layer.add_module(
        "output", Linear(config.intermediate_size, config.hidden_size)
    )
    feed_forward_layer.add_module("dropout", Dropout(config.hidden_dropout_prob))
    return feed_forward_layer


def get_attention_layer(config: PersformerConfig) -> Module:
    """
    Get the attention layer.
    """
    return AttentionFactory().build(config)  # type: ignore


def get_activation_function(activation_function: ActivationFunction) -> Module:
    """
    Get the activation function.
    """
    if activation_function is ActivationFunction.RELU:
        return nn.ReLU()
    elif activation_function is ActivationFunction.GELU:
        return nn.GELU()
    elif activation_function is ActivationFunction.SELU:
        return nn.SELU()
    elif activation_function is ActivationFunction.MISH:
        return nn.Mish()
    else:
        raise ValueError("Unknown activation function.")



from typing import Optional

import torch

from ..persformer_config import PersformerConfig
from .attention_base import AttentionBase

# Type aliases
from gdeep.utility.custom_types import Tensor


class InducedAttention(AttentionBase):
    """Class implementing the induced attention"""

    def __init__(self, config: PersformerConfig) -> None:
        super().__init__(config)
        raise NotImplementedError

    def forward(
        self, input: Tensor, attention_mask: Optional[Tensor] = None  # type: ignore
    ):
        raise NotImplementedError



"""Testing for UniformlySampledPoint."""
# License: GNU AGPLv3
import numpy as np
from .. import UniformlySampledPoint


def test_sample_points_uniformly():
    sample_points = UniformlySampledPoint([(2, 3), (1, 2)], n_samples=1000)

    print(sample_points().shape)
    assert sample_points().shape == (1000, 2)
    assert (sample_points().min(axis=0) >= np.array([2, 1])).all()
    assert (sample_points().min(axis=0) <= np.array([3, 2])).all()



from typing import Dict, Union

from torch import nn
import torchvision.models as models
from torch.optim import SGD
from torch.utils.data.sampler import SubsetRandomSampler

from gdeep.models import FFNet
from gdeep.data.datasets import DatasetBuilder, DataLoaderBuilder
from gdeep.data.preprocessors import ToTensorImage
from gdeep.search import HyperParameterOptimization, GiottoSummaryWriter, Benchmark


# parametric model with string value
class model2(nn.Module):
    def __init__(self, n_nodes="100"):
        super(model2, self).__init__()
        self.md = nn.Sequential(
            nn.Sequential(
                models.resnet18(pretrained=True), nn.Linear(1000, eval(n_nodes))
            ),
            nn.Linear(eval(n_nodes), 10),
        )

    def forward(self, x):
        return self.md(x)


models_dicts = []

model = model2()

temp_dict: Dict[str, Union[nn.Module, str]] = {"name": "resnet18", "model": model}

models_dicts.append(temp_dict)


# avoid having exposed parameters that wll not be searched on
class model_no_param(nn.Module):
    def __init__(self):
        super(model_no_param, self).__init__()
        self.mod = FFNet((3, 5, 5, 2))

    def forward(self, x):
        return self.mod(x)


model5 = model_no_param()
temp_dict = {"name": "ffnn", "model": model5}

models_dicts.append(temp_dict)


writer = GiottoSummaryWriter()

# download CIFAR10
bd = DatasetBuilder(name="CIFAR10")
ds_tr, _, _ = bd.build(download=True)

# Preprocessing
transformation = ToTensorImage([32, 32])
transformation.fit_to_dataset(ds_tr)  # this is useless for this transformation

transformed_ds_tr = transformation.attach_transform_to_dataset(ds_tr)

dataloaders_dicts = []


test_indices = [64 * 5 + x for x in range(32 * 3)]
train_indices = [x for x in range(32 * 2)]

dl = DataLoaderBuilder([transformed_ds_tr, transformed_ds_tr])
dl_tr, dl_val, _ = dl.build(  # type: ignore
    [
        {"batch_size": 32, "sampler": SubsetRandomSampler(train_indices)},  # type: ignore
        {"batch_size": 32, "sampler": SubsetRandomSampler(test_indices)},  # type: ignore
    ]
)


temp_dict = {"name": "CIFAR10_1000", "dataloaders": [dl_tr, dl_val]}  # type: ignore

dataloaders_dicts.append(temp_dict)

db = DatasetBuilder(name="DoubleTori")
ds_tr, ds_val, _ = db.build()

dl_tr, dl_ts, _ = DataLoaderBuilder([ds_tr, ds_val]).build([{"batch_size": 48}, {"batch_size": 32}])  # type: ignore

temp_dict = {"name": "double_tori", "dataloaders": [dl_tr, dl_ts]}  # type: ignore

dataloaders_dicts.append(temp_dict)


def test_benchmark():
    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise the benchmarking class. When we do not specify it, it will use KFold with 5 splits
    bench = Benchmark(models_dicts, dataloaders_dicts, loss_fn, writer)

    # start the benchmarking
    bench.start(SGD, 1, True, {"lr": 0.01}, {"batch_size": 23})


def test_benchmark2():
    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise the benchmarking class. When we do not specify it, it will use KFold with 5 splits
    bench = Benchmark(models_dicts, dataloaders_dicts, loss_fn, writer)

    # start the benchmarking
    bench.start(SGD, 2, False, {"lr": 0.01}, {"batch_size": 32}, n_accumulated_grads=2)



from dataclasses import dataclass
from typing import List, Any, Type, Optional, Dict

from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler  # noqa


@dataclass
class HPOConfig:
    """Config class for thee HPO parameters

    Args:
        optimizers:
            list of torch optimizers classes, not isntances
        n_epochs:
            number of training epochs
        cross_validation:
            whether or not to use cross-validation
        optimizers_params:
            dictionary of optimizers params
        dataloaders_params:
            dictionary of dataloaders parameters
        models_hyperparams:
            dictionary of model parameters
        lr_scheduler:
            torch learning rate schduler class
        schedulers_params:
            learning rate scheduler parameters
        profiling :
            whether or not you want to activate the
            profiler
        parallel_tpu:
            boolean value to run the computations
            on multiple TPUs
        n_accumulated_grads (int, default=0):
            number of accumulated gradients. It is
            considered only if a positive integer
        keep_training:
            bool flag to decide whether to continue
            training or not
        store_grad_layer_hist:
            flag to store the gradents of the layers in the
            tensorboard histograms
        writer_tag:
            tag to prepend to the ouput
            on tensorboard"""

    optimizers: List[Type[Optimizer]]
    n_epochs: int = 1
    cross_validation: bool = False
    optimizers_params: Optional[Dict[str, Any]] = None
    dataloaders_params: Optional[Dict[str, Any]] = None
    models_hyperparams: Optional[Dict[str, Any]] = None
    lr_scheduler: Optional[Type[_LRScheduler]] = None
    schedulers_params: Optional[Dict[str, Any]] = None
    profiling: bool = False
    parallel_tpu: bool = False
    keep_training: bool = False
    store_grad_layer_hist: bool = False
    n_accumulated_grads: int = 0
    writer_tag: str = ""

    def to_dict(self) -> Dict[str, Any]:
        """method to transform the config file into a dictionary"""
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}



import numpy as np  # type: ignore
import warnings


def generate_orbit(num_pts_per_orbit: int = 100, parameter: float = 1.0) -> np.ndarray:
    """Generate sequence of points of a dynamical system
    Non-parallel version

    Args:
        num_pts_per_orbit (int, optional):
            Number of points to generate.
            Defaults to 100.
        parameter (float, optional):
            Parameter of the dynamical system.
            Defaults to 1.0.

    Returns:
        np.ndarray:
            Array of sampled points of the dynamical system.
    """
    warnings.warn(
        "This methode is deprecated. Use the parallel\
                   version create_data.parallel_orbit.generate_\
                       orbit_parallel",
        DeprecationWarning,
    )
    try:
        assert parameter > 0
    except AssertionError:
        print("Parameter must be greater than 0")
    try:
        assert num_pts_per_orbit > 0
    except AssertionError:
        print("num_pts_per_orbit must be greater than 0")

    orbit = np.zeros([num_pts_per_orbit, 2])
    x_cur, y_cur = np.random.rand(), np.random.rand()  # current points
    for idx in range(num_pts_per_orbit):
        x_cur = (x_cur + parameter * y_cur * (1.0 - y_cur)) % 1
        y_cur = (y_cur + parameter * x_cur * (1.0 - x_cur)) % 1
        orbit[idx, :] = [x_cur, y_cur]
    return orbit



from .normalization import Normalization
from .tokenizer_translation import TokenizerTranslation
from .tokenizer_qa import TokenizerQA
from .to_tensor_image import ToTensorImage
from .tokenizer_text_classification import TokenizerTextClassification
from .filter_persistence_diagram_by_lifetime import FilterPersistenceDiagramByLifetime
from .filter_persistence_diagram_by_homology_dimension import (
    FilterPersistenceDiagramByHomologyDimension,
)
from .normalization_persistence_diagram import NormalizationPersistenceDiagram
from .min_max_normalization_persistence_diagram import MinMaxScalarPersistenceDiagram

__all__ = [
    "Normalization",
    "TokenizerTranslation",
    "TokenizerQA",
    "ToTensorImage",
    "TokenizerTextClassification",
    "NormalizationPersistenceDiagram",
    "FilterPersistenceDiagramByLifetime",
    "FilterPersistenceDiagramByHomologyDimension",
    "MinMaxScalarPersistenceDiagram",
]



from .benchmark import Benchmark, _benchmarking_param, TrainerConfig
from .hpo import HyperParameterOptimization, GiottoSummaryWriter
from ._utils import clean_up_files
from .hpo_config import HPOConfig

__all__ = [
    "Benchmark",
    "TrainerConfig",
    "HPOConfig",
    "HyperParameterOptimization",
    "GiottoSummaryWriter",
    "clean_up_files",
]



from .utils import (
    _are_compatible,
    save_model_and_optimizer,
    ensemble_wrapper,
    _inner_refactor_scalars,
    is_notebook,
    autoreload_if_notebook,
    KnownWarningSilencer,
)
from .constants import (
    ROOT_DIR,
    DEFAULT_DATA_DIR,
    DATASET_BUCKET_NAME,
    DEFAULT_DOWNLOAD_DIR,
    DATASET_BUCKET_NAME,
    DEFAULT_GRAPH_DIR,
    DEVICE,
)

from ._typing_utils import torch_transform, get_parameter_types, get_return_type

from .enum_types import (
    PoolerType,
    LayerNormStyle,
    AttentionType,
    ActivationFunction,
)

from . import optimization
from . import extended_persistence

__all__ = [
    "save_model_and_optimizer",
    "optimization",
    "extended_persistence",
    "ensemble_wrapper",
    "is_notebook",
    "autoreload_if_notebook",
    "ROOT_DIR",
    "DEVICE",
    "DEFAULT_DATA_DIR",
    "DATASET_BUCKET_NAME",
    "DEFAULT_DOWNLOAD_DIR",
    "DATASET_BUCKET_NAME",
    "DEFAULT_GRAPH_DIR",
    "get_return_type",
    "torch_transform",
    "get_parameter_types",
    "KnownWarningSilencer",
    "PoolerType",
    "LayerNormStyle",
    "AttentionType",
    "ActivationFunction",
]



import numpy as np


class UniformlySampledPoint:
    """Generates n_samples uniformely random points in a box
    specified in tuple_list

    Args:
        tuple_list (list): list of dimensionwise upper and lower bounds of box
        n_samples (int, optional): number of sample points. Defaults to 1000.
    """

    def __init__(self, tuple_list: list, n_samples: int = 1000):

        self._dim = len(tuple_list)
        try:
            for (left, right) in tuple_list:
                assert left <= right
        except AssertionError:
            raise ValueError("Tuples have have to be non-empty intervals")

        scale = np.array([[right - left for (left, right) in tuple_list]])
        translation = np.array([[left for (left, _) in tuple_list]])

        self._sample_points = np.random.rand(n_samples, self._dim) * scale + translation  # type: ignore

    def __call__(self):
        return self._sample_points

    def get_dim(self):
        """Returns dimension of sample point cloud

        Returns:
            int: dimension of point cloud
        """
        return self._dim
